from cognite.powerops.client._generated.data_classes._core import (
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    GraphQLList,
    ResourcesWrite,
    ResourcesWriteResult,
    PageInfo,
    TimeSeriesGraphQL,
    FileMetadataGraphQL,
    SequenceColumnGraphQL,
    SequenceGraphQL,
)
from ._alert import Alert, AlertFields, AlertGraphQL, AlertList, AlertTextFields, AlertWrite, AlertWriteList
from ._benchmarking_calculation_input import BenchmarkingCalculationInput, BenchmarkingCalculationInputFields, BenchmarkingCalculationInputGraphQL, BenchmarkingCalculationInputList, BenchmarkingCalculationInputTextFields, BenchmarkingCalculationInputWrite, BenchmarkingCalculationInputWriteList
from ._benchmarking_calculation_output import BenchmarkingCalculationOutput, BenchmarkingCalculationOutputFields, BenchmarkingCalculationOutputGraphQL, BenchmarkingCalculationOutputList, BenchmarkingCalculationOutputTextFields, BenchmarkingCalculationOutputWrite, BenchmarkingCalculationOutputWriteList
from ._benchmarking_configuration_day_ahead import BenchmarkingConfigurationDayAhead, BenchmarkingConfigurationDayAheadFields, BenchmarkingConfigurationDayAheadGraphQL, BenchmarkingConfigurationDayAheadList, BenchmarkingConfigurationDayAheadTextFields, BenchmarkingConfigurationDayAheadWrite, BenchmarkingConfigurationDayAheadWriteList
from ._benchmarking_production_obligation_day_ahead import BenchmarkingProductionObligationDayAhead, BenchmarkingProductionObligationDayAheadFields, BenchmarkingProductionObligationDayAheadGraphQL, BenchmarkingProductionObligationDayAheadList, BenchmarkingProductionObligationDayAheadTextFields, BenchmarkingProductionObligationDayAheadWrite, BenchmarkingProductionObligationDayAheadWriteList
from ._benchmarking_result_day_ahead import BenchmarkingResultDayAhead, BenchmarkingResultDayAheadFields, BenchmarkingResultDayAheadGraphQL, BenchmarkingResultDayAheadList, BenchmarkingResultDayAheadTextFields, BenchmarkingResultDayAheadWrite, BenchmarkingResultDayAheadWriteList
from ._benchmarking_shop_case import BenchmarkingShopCase, BenchmarkingShopCaseFields, BenchmarkingShopCaseGraphQL, BenchmarkingShopCaseList, BenchmarkingShopCaseTextFields, BenchmarkingShopCaseWrite, BenchmarkingShopCaseWriteList
from ._benchmarking_task_dispatcher_input_day_ahead import BenchmarkingTaskDispatcherInputDayAhead, BenchmarkingTaskDispatcherInputDayAheadFields, BenchmarkingTaskDispatcherInputDayAheadGraphQL, BenchmarkingTaskDispatcherInputDayAheadList, BenchmarkingTaskDispatcherInputDayAheadTextFields, BenchmarkingTaskDispatcherInputDayAheadWrite, BenchmarkingTaskDispatcherInputDayAheadWriteList
from ._benchmarking_task_dispatcher_output_day_ahead import BenchmarkingTaskDispatcherOutputDayAhead, BenchmarkingTaskDispatcherOutputDayAheadFields, BenchmarkingTaskDispatcherOutputDayAheadGraphQL, BenchmarkingTaskDispatcherOutputDayAheadList, BenchmarkingTaskDispatcherOutputDayAheadTextFields, BenchmarkingTaskDispatcherOutputDayAheadWrite, BenchmarkingTaskDispatcherOutputDayAheadWriteList
from ._bid_configuration_day_ahead import BidConfigurationDayAhead, BidConfigurationDayAheadFields, BidConfigurationDayAheadGraphQL, BidConfigurationDayAheadList, BidConfigurationDayAheadTextFields, BidConfigurationDayAheadWrite, BidConfigurationDayAheadWriteList
from ._bid_document import BidDocument, BidDocumentFields, BidDocumentGraphQL, BidDocumentList, BidDocumentTextFields, BidDocumentWrite, BidDocumentWriteList
from ._bid_document_afrr import BidDocumentAFRR, BidDocumentAFRRFields, BidDocumentAFRRGraphQL, BidDocumentAFRRList, BidDocumentAFRRTextFields, BidDocumentAFRRWrite, BidDocumentAFRRWriteList
from ._bid_document_day_ahead import BidDocumentDayAhead, BidDocumentDayAheadFields, BidDocumentDayAheadGraphQL, BidDocumentDayAheadList, BidDocumentDayAheadTextFields, BidDocumentDayAheadWrite, BidDocumentDayAheadWriteList
from ._bid_matrix import BidMatrix, BidMatrixFields, BidMatrixGraphQL, BidMatrixList, BidMatrixTextFields, BidMatrixWrite, BidMatrixWriteList
from ._bid_matrix_information import BidMatrixInformation, BidMatrixInformationFields, BidMatrixInformationGraphQL, BidMatrixInformationList, BidMatrixInformationTextFields, BidMatrixInformationWrite, BidMatrixInformationWriteList
from ._bid_row import BidRow, BidRowFields, BidRowGraphQL, BidRowList, BidRowTextFields, BidRowWrite, BidRowWriteList
from ._data_set_configuration import DataSetConfiguration, DataSetConfigurationFields, DataSetConfigurationGraphQL, DataSetConfigurationList, DataSetConfigurationTextFields, DataSetConfigurationWrite, DataSetConfigurationWriteList
from ._date_specification import DateSpecification, DateSpecificationFields, DateSpecificationGraphQL, DateSpecificationList, DateSpecificationTextFields, DateSpecificationWrite, DateSpecificationWriteList
from ._function_input import FunctionInput, FunctionInputFields, FunctionInputGraphQL, FunctionInputList, FunctionInputTextFields, FunctionInputWrite, FunctionInputWriteList
from ._function_output import FunctionOutput, FunctionOutputFields, FunctionOutputGraphQL, FunctionOutputList, FunctionOutputTextFields, FunctionOutputWrite, FunctionOutputWriteList
from ._generator import Generator, GeneratorFields, GeneratorGraphQL, GeneratorList, GeneratorTextFields, GeneratorWrite, GeneratorWriteList
from ._generator_efficiency_curve import GeneratorEfficiencyCurve, GeneratorEfficiencyCurveFields, GeneratorEfficiencyCurveGraphQL, GeneratorEfficiencyCurveList, GeneratorEfficiencyCurveTextFields, GeneratorEfficiencyCurveWrite, GeneratorEfficiencyCurveWriteList
from ._market_configuration import MarketConfiguration, MarketConfigurationFields, MarketConfigurationGraphQL, MarketConfigurationList, MarketConfigurationTextFields, MarketConfigurationWrite, MarketConfigurationWriteList
from ._multi_scenario_partial_bid_matrix_calculation_input import MultiScenarioPartialBidMatrixCalculationInput, MultiScenarioPartialBidMatrixCalculationInputFields, MultiScenarioPartialBidMatrixCalculationInputGraphQL, MultiScenarioPartialBidMatrixCalculationInputList, MultiScenarioPartialBidMatrixCalculationInputTextFields, MultiScenarioPartialBidMatrixCalculationInputWrite, MultiScenarioPartialBidMatrixCalculationInputWriteList
from ._partial_bid_configuration import PartialBidConfiguration, PartialBidConfigurationFields, PartialBidConfigurationGraphQL, PartialBidConfigurationList, PartialBidConfigurationTextFields, PartialBidConfigurationWrite, PartialBidConfigurationWriteList
from ._partial_bid_matrix_calculation_input import PartialBidMatrixCalculationInput, PartialBidMatrixCalculationInputFields, PartialBidMatrixCalculationInputGraphQL, PartialBidMatrixCalculationInputList, PartialBidMatrixCalculationInputTextFields, PartialBidMatrixCalculationInputWrite, PartialBidMatrixCalculationInputWriteList
from ._partial_bid_matrix_calculation_output import PartialBidMatrixCalculationOutput, PartialBidMatrixCalculationOutputFields, PartialBidMatrixCalculationOutputGraphQL, PartialBidMatrixCalculationOutputList, PartialBidMatrixCalculationOutputTextFields, PartialBidMatrixCalculationOutputWrite, PartialBidMatrixCalculationOutputWriteList
from ._partial_bid_matrix_information import PartialBidMatrixInformation, PartialBidMatrixInformationFields, PartialBidMatrixInformationGraphQL, PartialBidMatrixInformationList, PartialBidMatrixInformationTextFields, PartialBidMatrixInformationWrite, PartialBidMatrixInformationWriteList
from ._partial_bid_matrix_information_with_scenarios import PartialBidMatrixInformationWithScenarios, PartialBidMatrixInformationWithScenariosFields, PartialBidMatrixInformationWithScenariosGraphQL, PartialBidMatrixInformationWithScenariosList, PartialBidMatrixInformationWithScenariosTextFields, PartialBidMatrixInformationWithScenariosWrite, PartialBidMatrixInformationWithScenariosWriteList
from ._plant import Plant, PlantFields, PlantGraphQL, PlantList, PlantTextFields, PlantWrite, PlantWriteList
from ._plant_information import PlantInformation, PlantInformationFields, PlantInformationGraphQL, PlantInformationList, PlantInformationTextFields, PlantInformationWrite, PlantInformationWriteList
from ._plant_water_value_based import PlantWaterValueBased, PlantWaterValueBasedFields, PlantWaterValueBasedGraphQL, PlantWaterValueBasedList, PlantWaterValueBasedTextFields, PlantWaterValueBasedWrite, PlantWaterValueBasedWriteList
from ._power_asset import PowerAsset, PowerAssetFields, PowerAssetGraphQL, PowerAssetList, PowerAssetTextFields, PowerAssetWrite, PowerAssetWriteList
from ._price_area import PriceArea, PriceAreaFields, PriceAreaGraphQL, PriceAreaList, PriceAreaTextFields, PriceAreaWrite, PriceAreaWriteList
from ._price_area_afrr import PriceAreaAFRR, PriceAreaAFRRFields, PriceAreaAFRRGraphQL, PriceAreaAFRRList, PriceAreaAFRRTextFields, PriceAreaAFRRWrite, PriceAreaAFRRWriteList
from ._price_area_day_ahead import PriceAreaDayAhead, PriceAreaDayAheadFields, PriceAreaDayAheadGraphQL, PriceAreaDayAheadList, PriceAreaDayAheadTextFields, PriceAreaDayAheadWrite, PriceAreaDayAheadWriteList
from ._price_area_information import PriceAreaInformation, PriceAreaInformationFields, PriceAreaInformationGraphQL, PriceAreaInformationList, PriceAreaInformationTextFields, PriceAreaInformationWrite, PriceAreaInformationWriteList
from ._price_production import PriceProduction, PriceProductionFields, PriceProductionGraphQL, PriceProductionList, PriceProductionTextFields, PriceProductionWrite, PriceProductionWriteList
from ._shop_attribute_mapping import ShopAttributeMapping, ShopAttributeMappingFields, ShopAttributeMappingGraphQL, ShopAttributeMappingList, ShopAttributeMappingTextFields, ShopAttributeMappingWrite, ShopAttributeMappingWriteList
from ._shop_based_partial_bid_configuration import ShopBasedPartialBidConfiguration, ShopBasedPartialBidConfigurationFields, ShopBasedPartialBidConfigurationGraphQL, ShopBasedPartialBidConfigurationList, ShopBasedPartialBidConfigurationTextFields, ShopBasedPartialBidConfigurationWrite, ShopBasedPartialBidConfigurationWriteList
from ._shop_case import ShopCase, ShopCaseFields, ShopCaseGraphQL, ShopCaseList, ShopCaseTextFields, ShopCaseWrite, ShopCaseWriteList
from ._shop_commands import ShopCommands, ShopCommandsFields, ShopCommandsGraphQL, ShopCommandsList, ShopCommandsTextFields, ShopCommandsWrite, ShopCommandsWriteList
from ._shop_file import ShopFile, ShopFileFields, ShopFileGraphQL, ShopFileList, ShopFileTextFields, ShopFileWrite, ShopFileWriteList
from ._shop_model import ShopModel, ShopModelFields, ShopModelGraphQL, ShopModelList, ShopModelTextFields, ShopModelWrite, ShopModelWriteList
from ._shop_model_with_assets import ShopModelWithAssets, ShopModelWithAssetsFields, ShopModelWithAssetsGraphQL, ShopModelWithAssetsList, ShopModelWithAssetsTextFields, ShopModelWithAssetsWrite, ShopModelWithAssetsWriteList
from ._shop_output_time_series_definition import ShopOutputTimeSeriesDefinition, ShopOutputTimeSeriesDefinitionFields, ShopOutputTimeSeriesDefinitionGraphQL, ShopOutputTimeSeriesDefinitionList, ShopOutputTimeSeriesDefinitionTextFields, ShopOutputTimeSeriesDefinitionWrite, ShopOutputTimeSeriesDefinitionWriteList
from ._shop_penalty_report import ShopPenaltyReport, ShopPenaltyReportFields, ShopPenaltyReportGraphQL, ShopPenaltyReportList, ShopPenaltyReportTextFields, ShopPenaltyReportWrite, ShopPenaltyReportWriteList
from ._shop_preprocessor_input import ShopPreprocessorInput, ShopPreprocessorInputFields, ShopPreprocessorInputGraphQL, ShopPreprocessorInputList, ShopPreprocessorInputTextFields, ShopPreprocessorInputWrite, ShopPreprocessorInputWriteList
from ._shop_preprocessor_output import ShopPreprocessorOutput, ShopPreprocessorOutputFields, ShopPreprocessorOutputGraphQL, ShopPreprocessorOutputList, ShopPreprocessorOutputTextFields, ShopPreprocessorOutputWrite, ShopPreprocessorOutputWriteList
from ._shop_result import ShopResult, ShopResultFields, ShopResultGraphQL, ShopResultList, ShopResultTextFields, ShopResultWrite, ShopResultWriteList
from ._shop_scenario import ShopScenario, ShopScenarioFields, ShopScenarioGraphQL, ShopScenarioList, ShopScenarioTextFields, ShopScenarioWrite, ShopScenarioWriteList
from ._shop_scenario_set import ShopScenarioSet, ShopScenarioSetFields, ShopScenarioSetGraphQL, ShopScenarioSetList, ShopScenarioSetTextFields, ShopScenarioSetWrite, ShopScenarioSetWriteList
from ._shop_time_resolution import ShopTimeResolution, ShopTimeResolutionFields, ShopTimeResolutionGraphQL, ShopTimeResolutionList, ShopTimeResolutionTextFields, ShopTimeResolutionWrite, ShopTimeResolutionWriteList
from ._shop_time_series import ShopTimeSeries, ShopTimeSeriesFields, ShopTimeSeriesGraphQL, ShopTimeSeriesList, ShopTimeSeriesTextFields, ShopTimeSeriesWrite, ShopTimeSeriesWriteList
from ._shop_trigger_input import ShopTriggerInput, ShopTriggerInputFields, ShopTriggerInputGraphQL, ShopTriggerInputList, ShopTriggerInputTextFields, ShopTriggerInputWrite, ShopTriggerInputWriteList
from ._shop_trigger_output import ShopTriggerOutput, ShopTriggerOutputFields, ShopTriggerOutputGraphQL, ShopTriggerOutputList, ShopTriggerOutputTextFields, ShopTriggerOutputWrite, ShopTriggerOutputWriteList
from ._task_dispatcher_input import TaskDispatcherInput, TaskDispatcherInputFields, TaskDispatcherInputGraphQL, TaskDispatcherInputList, TaskDispatcherInputTextFields, TaskDispatcherInputWrite, TaskDispatcherInputWriteList
from ._task_dispatcher_output import TaskDispatcherOutput, TaskDispatcherOutputFields, TaskDispatcherOutputGraphQL, TaskDispatcherOutputList, TaskDispatcherOutputTextFields, TaskDispatcherOutputWrite, TaskDispatcherOutputWriteList
from ._total_bid_matrix_calculation_input import TotalBidMatrixCalculationInput, TotalBidMatrixCalculationInputFields, TotalBidMatrixCalculationInputGraphQL, TotalBidMatrixCalculationInputList, TotalBidMatrixCalculationInputTextFields, TotalBidMatrixCalculationInputWrite, TotalBidMatrixCalculationInputWriteList
from ._total_bid_matrix_calculation_output import TotalBidMatrixCalculationOutput, TotalBidMatrixCalculationOutputFields, TotalBidMatrixCalculationOutputGraphQL, TotalBidMatrixCalculationOutputList, TotalBidMatrixCalculationOutputTextFields, TotalBidMatrixCalculationOutputWrite, TotalBidMatrixCalculationOutputWriteList
from ._turbine_efficiency_curve import TurbineEfficiencyCurve, TurbineEfficiencyCurveFields, TurbineEfficiencyCurveGraphQL, TurbineEfficiencyCurveList, TurbineEfficiencyCurveTextFields, TurbineEfficiencyCurveWrite, TurbineEfficiencyCurveWriteList
from ._water_value_based_partial_bid_configuration import WaterValueBasedPartialBidConfiguration, WaterValueBasedPartialBidConfigurationFields, WaterValueBasedPartialBidConfigurationGraphQL, WaterValueBasedPartialBidConfigurationList, WaterValueBasedPartialBidConfigurationTextFields, WaterValueBasedPartialBidConfigurationWrite, WaterValueBasedPartialBidConfigurationWriteList
from ._water_value_based_partial_bid_matrix_calculation_input import WaterValueBasedPartialBidMatrixCalculationInput, WaterValueBasedPartialBidMatrixCalculationInputFields, WaterValueBasedPartialBidMatrixCalculationInputGraphQL, WaterValueBasedPartialBidMatrixCalculationInputList, WaterValueBasedPartialBidMatrixCalculationInputTextFields, WaterValueBasedPartialBidMatrixCalculationInputWrite, WaterValueBasedPartialBidMatrixCalculationInputWriteList
from ._watercourse import Watercourse, WatercourseFields, WatercourseGraphQL, WatercourseList, WatercourseTextFields, WatercourseWrite, WatercourseWriteList
BenchmarkingCalculationInput.model_rebuild()
BenchmarkingCalculationInputGraphQL.model_rebuild()
BenchmarkingCalculationInputWrite.model_rebuild()
BenchmarkingCalculationOutput.model_rebuild()
BenchmarkingCalculationOutputGraphQL.model_rebuild()
BenchmarkingCalculationOutputWrite.model_rebuild()
BenchmarkingConfigurationDayAhead.model_rebuild()
BenchmarkingConfigurationDayAheadGraphQL.model_rebuild()
BenchmarkingConfigurationDayAheadWrite.model_rebuild()
BenchmarkingProductionObligationDayAhead.model_rebuild()
BenchmarkingProductionObligationDayAheadGraphQL.model_rebuild()
BenchmarkingProductionObligationDayAheadWrite.model_rebuild()
BenchmarkingResultDayAhead.model_rebuild()
BenchmarkingResultDayAheadGraphQL.model_rebuild()
BenchmarkingResultDayAheadWrite.model_rebuild()
BenchmarkingShopCase.model_rebuild()
BenchmarkingShopCaseGraphQL.model_rebuild()
BenchmarkingShopCaseWrite.model_rebuild()
BenchmarkingTaskDispatcherInputDayAhead.model_rebuild()
BenchmarkingTaskDispatcherInputDayAheadGraphQL.model_rebuild()
BenchmarkingTaskDispatcherInputDayAheadWrite.model_rebuild()
BenchmarkingTaskDispatcherOutputDayAhead.model_rebuild()
BenchmarkingTaskDispatcherOutputDayAheadGraphQL.model_rebuild()
BenchmarkingTaskDispatcherOutputDayAheadWrite.model_rebuild()
BidConfigurationDayAhead.model_rebuild()
BidConfigurationDayAheadGraphQL.model_rebuild()
BidConfigurationDayAheadWrite.model_rebuild()
BidDocument.model_rebuild()
BidDocumentGraphQL.model_rebuild()
BidDocumentWrite.model_rebuild()
BidDocumentAFRR.model_rebuild()
BidDocumentAFRRGraphQL.model_rebuild()
BidDocumentAFRRWrite.model_rebuild()
BidDocumentDayAhead.model_rebuild()
BidDocumentDayAheadGraphQL.model_rebuild()
BidDocumentDayAheadWrite.model_rebuild()
BidMatrixInformation.model_rebuild()
BidMatrixInformationGraphQL.model_rebuild()
BidMatrixInformationWrite.model_rebuild()
BidRow.model_rebuild()
BidRowGraphQL.model_rebuild()
BidRowWrite.model_rebuild()
FunctionOutput.model_rebuild()
FunctionOutputGraphQL.model_rebuild()
FunctionOutputWrite.model_rebuild()
Generator.model_rebuild()
GeneratorGraphQL.model_rebuild()
GeneratorWrite.model_rebuild()
MultiScenarioPartialBidMatrixCalculationInput.model_rebuild()
MultiScenarioPartialBidMatrixCalculationInputGraphQL.model_rebuild()
MultiScenarioPartialBidMatrixCalculationInputWrite.model_rebuild()
PartialBidConfiguration.model_rebuild()
PartialBidConfigurationGraphQL.model_rebuild()
PartialBidConfigurationWrite.model_rebuild()
PartialBidMatrixCalculationInput.model_rebuild()
PartialBidMatrixCalculationInputGraphQL.model_rebuild()
PartialBidMatrixCalculationInputWrite.model_rebuild()
PartialBidMatrixCalculationOutput.model_rebuild()
PartialBidMatrixCalculationOutputGraphQL.model_rebuild()
PartialBidMatrixCalculationOutputWrite.model_rebuild()
PartialBidMatrixInformation.model_rebuild()
PartialBidMatrixInformationGraphQL.model_rebuild()
PartialBidMatrixInformationWrite.model_rebuild()
PartialBidMatrixInformationWithScenarios.model_rebuild()
PartialBidMatrixInformationWithScenariosGraphQL.model_rebuild()
PartialBidMatrixInformationWithScenariosWrite.model_rebuild()
PlantInformation.model_rebuild()
PlantInformationGraphQL.model_rebuild()
PlantInformationWrite.model_rebuild()
PlantWaterValueBased.model_rebuild()
PlantWaterValueBasedGraphQL.model_rebuild()
PlantWaterValueBasedWrite.model_rebuild()
PriceAreaAFRR.model_rebuild()
PriceAreaAFRRGraphQL.model_rebuild()
PriceAreaAFRRWrite.model_rebuild()
PriceAreaDayAhead.model_rebuild()
PriceAreaDayAheadGraphQL.model_rebuild()
PriceAreaDayAheadWrite.model_rebuild()
PriceAreaInformation.model_rebuild()
PriceAreaInformationGraphQL.model_rebuild()
PriceAreaInformationWrite.model_rebuild()
PriceProduction.model_rebuild()
PriceProductionGraphQL.model_rebuild()
PriceProductionWrite.model_rebuild()
ShopAttributeMapping.model_rebuild()
ShopAttributeMappingGraphQL.model_rebuild()
ShopAttributeMappingWrite.model_rebuild()
ShopBasedPartialBidConfiguration.model_rebuild()
ShopBasedPartialBidConfigurationGraphQL.model_rebuild()
ShopBasedPartialBidConfigurationWrite.model_rebuild()
ShopCase.model_rebuild()
ShopCaseGraphQL.model_rebuild()
ShopCaseWrite.model_rebuild()
ShopModel.model_rebuild()
ShopModelGraphQL.model_rebuild()
ShopModelWrite.model_rebuild()
ShopModelWithAssets.model_rebuild()
ShopModelWithAssetsGraphQL.model_rebuild()
ShopModelWithAssetsWrite.model_rebuild()
ShopPreprocessorInput.model_rebuild()
ShopPreprocessorInputGraphQL.model_rebuild()
ShopPreprocessorInputWrite.model_rebuild()
ShopPreprocessorOutput.model_rebuild()
ShopPreprocessorOutputGraphQL.model_rebuild()
ShopPreprocessorOutputWrite.model_rebuild()
ShopResult.model_rebuild()
ShopResultGraphQL.model_rebuild()
ShopResultWrite.model_rebuild()
ShopScenario.model_rebuild()
ShopScenarioGraphQL.model_rebuild()
ShopScenarioWrite.model_rebuild()
ShopScenarioSet.model_rebuild()
ShopScenarioSetGraphQL.model_rebuild()
ShopScenarioSetWrite.model_rebuild()
ShopTimeSeries.model_rebuild()
ShopTimeSeriesGraphQL.model_rebuild()
ShopTimeSeriesWrite.model_rebuild()
ShopTriggerInput.model_rebuild()
ShopTriggerInputGraphQL.model_rebuild()
ShopTriggerInputWrite.model_rebuild()
ShopTriggerOutput.model_rebuild()
ShopTriggerOutputGraphQL.model_rebuild()
ShopTriggerOutputWrite.model_rebuild()
TaskDispatcherInput.model_rebuild()
TaskDispatcherInputGraphQL.model_rebuild()
TaskDispatcherInputWrite.model_rebuild()
TaskDispatcherOutput.model_rebuild()
TaskDispatcherOutputGraphQL.model_rebuild()
TaskDispatcherOutputWrite.model_rebuild()
TotalBidMatrixCalculationInput.model_rebuild()
TotalBidMatrixCalculationInputGraphQL.model_rebuild()
TotalBidMatrixCalculationInputWrite.model_rebuild()
TotalBidMatrixCalculationOutput.model_rebuild()
TotalBidMatrixCalculationOutputGraphQL.model_rebuild()
TotalBidMatrixCalculationOutputWrite.model_rebuild()
WaterValueBasedPartialBidConfiguration.model_rebuild()
WaterValueBasedPartialBidConfigurationGraphQL.model_rebuild()
WaterValueBasedPartialBidConfigurationWrite.model_rebuild()
WaterValueBasedPartialBidMatrixCalculationInput.model_rebuild()
WaterValueBasedPartialBidMatrixCalculationInputGraphQL.model_rebuild()
WaterValueBasedPartialBidMatrixCalculationInputWrite.model_rebuild()


__all__ = [
    "DataRecord",
    "DataRecordGraphQL",
    "DataRecordWrite",
    "ResourcesWrite",
    "DomainModel",
    "DomainModelCore",
    "DomainModelWrite",
    "DomainModelList",
    "DomainRelationWrite",
    "GraphQLCore",
    "GraphQLList",
    "ResourcesWriteResult",
    "PageInfo",
    "TimeSeriesGraphQL",
    "FileMetadataGraphQL",
    "SequenceColumnGraphQL",
    "SequenceGraphQL",
    "Alert",
    "AlertGraphQL",
    "AlertWrite",
    "AlertList",
    "AlertWriteList",
    "AlertFields",
    "AlertTextFields",
    "BenchmarkingCalculationInput",
    "BenchmarkingCalculationInputGraphQL",
    "BenchmarkingCalculationInputWrite",
    "BenchmarkingCalculationInputList",
    "BenchmarkingCalculationInputWriteList",
    "BenchmarkingCalculationInputFields",
    "BenchmarkingCalculationInputTextFields",
    "BenchmarkingCalculationOutput",
    "BenchmarkingCalculationOutputGraphQL",
    "BenchmarkingCalculationOutputWrite",
    "BenchmarkingCalculationOutputList",
    "BenchmarkingCalculationOutputWriteList",
    "BenchmarkingCalculationOutputFields",
    "BenchmarkingCalculationOutputTextFields",
    "BenchmarkingConfigurationDayAhead",
    "BenchmarkingConfigurationDayAheadGraphQL",
    "BenchmarkingConfigurationDayAheadWrite",
    "BenchmarkingConfigurationDayAheadList",
    "BenchmarkingConfigurationDayAheadWriteList",
    "BenchmarkingConfigurationDayAheadFields",
    "BenchmarkingConfigurationDayAheadTextFields",
    "BenchmarkingProductionObligationDayAhead",
    "BenchmarkingProductionObligationDayAheadGraphQL",
    "BenchmarkingProductionObligationDayAheadWrite",
    "BenchmarkingProductionObligationDayAheadList",
    "BenchmarkingProductionObligationDayAheadWriteList",
    "BenchmarkingProductionObligationDayAheadFields",
    "BenchmarkingProductionObligationDayAheadTextFields",
    "BenchmarkingResultDayAhead",
    "BenchmarkingResultDayAheadGraphQL",
    "BenchmarkingResultDayAheadWrite",
    "BenchmarkingResultDayAheadList",
    "BenchmarkingResultDayAheadWriteList",
    "BenchmarkingResultDayAheadFields",
    "BenchmarkingResultDayAheadTextFields",
    "BenchmarkingShopCase",
    "BenchmarkingShopCaseGraphQL",
    "BenchmarkingShopCaseWrite",
    "BenchmarkingShopCaseList",
    "BenchmarkingShopCaseWriteList",
    "BenchmarkingShopCaseFields",
    "BenchmarkingShopCaseTextFields",
    "BenchmarkingTaskDispatcherInputDayAhead",
    "BenchmarkingTaskDispatcherInputDayAheadGraphQL",
    "BenchmarkingTaskDispatcherInputDayAheadWrite",
    "BenchmarkingTaskDispatcherInputDayAheadList",
    "BenchmarkingTaskDispatcherInputDayAheadWriteList",
    "BenchmarkingTaskDispatcherInputDayAheadFields",
    "BenchmarkingTaskDispatcherInputDayAheadTextFields",
    "BenchmarkingTaskDispatcherOutputDayAhead",
    "BenchmarkingTaskDispatcherOutputDayAheadGraphQL",
    "BenchmarkingTaskDispatcherOutputDayAheadWrite",
    "BenchmarkingTaskDispatcherOutputDayAheadList",
    "BenchmarkingTaskDispatcherOutputDayAheadWriteList",
    "BenchmarkingTaskDispatcherOutputDayAheadFields",
    "BenchmarkingTaskDispatcherOutputDayAheadTextFields",
    "BidConfigurationDayAhead",
    "BidConfigurationDayAheadGraphQL",
    "BidConfigurationDayAheadWrite",
    "BidConfigurationDayAheadList",
    "BidConfigurationDayAheadWriteList",
    "BidConfigurationDayAheadFields",
    "BidConfigurationDayAheadTextFields",
    "BidDocument",
    "BidDocumentGraphQL",
    "BidDocumentWrite",
    "BidDocumentList",
    "BidDocumentWriteList",
    "BidDocumentFields",
    "BidDocumentTextFields",
    "BidDocumentAFRR",
    "BidDocumentAFRRGraphQL",
    "BidDocumentAFRRWrite",
    "BidDocumentAFRRList",
    "BidDocumentAFRRWriteList",
    "BidDocumentAFRRFields",
    "BidDocumentAFRRTextFields",
    "BidDocumentDayAhead",
    "BidDocumentDayAheadGraphQL",
    "BidDocumentDayAheadWrite",
    "BidDocumentDayAheadList",
    "BidDocumentDayAheadWriteList",
    "BidDocumentDayAheadFields",
    "BidDocumentDayAheadTextFields",
    "BidMatrix",
    "BidMatrixGraphQL",
    "BidMatrixWrite",
    "BidMatrixList",
    "BidMatrixWriteList",
    "BidMatrixFields",
    "BidMatrixTextFields",
    "BidMatrixInformation",
    "BidMatrixInformationGraphQL",
    "BidMatrixInformationWrite",
    "BidMatrixInformationList",
    "BidMatrixInformationWriteList",
    "BidMatrixInformationFields",
    "BidMatrixInformationTextFields",
    "BidRow",
    "BidRowGraphQL",
    "BidRowWrite",
    "BidRowList",
    "BidRowWriteList",
    "BidRowFields",
    "BidRowTextFields",
    "DataSetConfiguration",
    "DataSetConfigurationGraphQL",
    "DataSetConfigurationWrite",
    "DataSetConfigurationList",
    "DataSetConfigurationWriteList",
    "DataSetConfigurationFields",
    "DataSetConfigurationTextFields",
    "DateSpecification",
    "DateSpecificationGraphQL",
    "DateSpecificationWrite",
    "DateSpecificationList",
    "DateSpecificationWriteList",
    "DateSpecificationFields",
    "DateSpecificationTextFields",
    "FunctionInput",
    "FunctionInputGraphQL",
    "FunctionInputWrite",
    "FunctionInputList",
    "FunctionInputWriteList",
    "FunctionInputFields",
    "FunctionInputTextFields",
    "FunctionOutput",
    "FunctionOutputGraphQL",
    "FunctionOutputWrite",
    "FunctionOutputList",
    "FunctionOutputWriteList",
    "FunctionOutputFields",
    "FunctionOutputTextFields",
    "Generator",
    "GeneratorGraphQL",
    "GeneratorWrite",
    "GeneratorList",
    "GeneratorWriteList",
    "GeneratorFields",
    "GeneratorTextFields",
    "GeneratorEfficiencyCurve",
    "GeneratorEfficiencyCurveGraphQL",
    "GeneratorEfficiencyCurveWrite",
    "GeneratorEfficiencyCurveList",
    "GeneratorEfficiencyCurveWriteList",
    "GeneratorEfficiencyCurveFields",
    "GeneratorEfficiencyCurveTextFields",
    "MarketConfiguration",
    "MarketConfigurationGraphQL",
    "MarketConfigurationWrite",
    "MarketConfigurationList",
    "MarketConfigurationWriteList",
    "MarketConfigurationFields",
    "MarketConfigurationTextFields",
    "MultiScenarioPartialBidMatrixCalculationInput",
    "MultiScenarioPartialBidMatrixCalculationInputGraphQL",
    "MultiScenarioPartialBidMatrixCalculationInputWrite",
    "MultiScenarioPartialBidMatrixCalculationInputList",
    "MultiScenarioPartialBidMatrixCalculationInputWriteList",
    "MultiScenarioPartialBidMatrixCalculationInputFields",
    "MultiScenarioPartialBidMatrixCalculationInputTextFields",
    "PartialBidConfiguration",
    "PartialBidConfigurationGraphQL",
    "PartialBidConfigurationWrite",
    "PartialBidConfigurationList",
    "PartialBidConfigurationWriteList",
    "PartialBidConfigurationFields",
    "PartialBidConfigurationTextFields",
    "PartialBidMatrixCalculationInput",
    "PartialBidMatrixCalculationInputGraphQL",
    "PartialBidMatrixCalculationInputWrite",
    "PartialBidMatrixCalculationInputList",
    "PartialBidMatrixCalculationInputWriteList",
    "PartialBidMatrixCalculationInputFields",
    "PartialBidMatrixCalculationInputTextFields",
    "PartialBidMatrixCalculationOutput",
    "PartialBidMatrixCalculationOutputGraphQL",
    "PartialBidMatrixCalculationOutputWrite",
    "PartialBidMatrixCalculationOutputList",
    "PartialBidMatrixCalculationOutputWriteList",
    "PartialBidMatrixCalculationOutputFields",
    "PartialBidMatrixCalculationOutputTextFields",
    "PartialBidMatrixInformation",
    "PartialBidMatrixInformationGraphQL",
    "PartialBidMatrixInformationWrite",
    "PartialBidMatrixInformationList",
    "PartialBidMatrixInformationWriteList",
    "PartialBidMatrixInformationFields",
    "PartialBidMatrixInformationTextFields",
    "PartialBidMatrixInformationWithScenarios",
    "PartialBidMatrixInformationWithScenariosGraphQL",
    "PartialBidMatrixInformationWithScenariosWrite",
    "PartialBidMatrixInformationWithScenariosList",
    "PartialBidMatrixInformationWithScenariosWriteList",
    "PartialBidMatrixInformationWithScenariosFields",
    "PartialBidMatrixInformationWithScenariosTextFields",
    "Plant",
    "PlantGraphQL",
    "PlantWrite",
    "PlantList",
    "PlantWriteList",
    "PlantFields",
    "PlantTextFields",
    "PlantInformation",
    "PlantInformationGraphQL",
    "PlantInformationWrite",
    "PlantInformationList",
    "PlantInformationWriteList",
    "PlantInformationFields",
    "PlantInformationTextFields",
    "PlantWaterValueBased",
    "PlantWaterValueBasedGraphQL",
    "PlantWaterValueBasedWrite",
    "PlantWaterValueBasedList",
    "PlantWaterValueBasedWriteList",
    "PlantWaterValueBasedFields",
    "PlantWaterValueBasedTextFields",
    "PowerAsset",
    "PowerAssetGraphQL",
    "PowerAssetWrite",
    "PowerAssetList",
    "PowerAssetWriteList",
    "PowerAssetFields",
    "PowerAssetTextFields",
    "PriceArea",
    "PriceAreaGraphQL",
    "PriceAreaWrite",
    "PriceAreaList",
    "PriceAreaWriteList",
    "PriceAreaFields",
    "PriceAreaTextFields",
    "PriceAreaAFRR",
    "PriceAreaAFRRGraphQL",
    "PriceAreaAFRRWrite",
    "PriceAreaAFRRList",
    "PriceAreaAFRRWriteList",
    "PriceAreaAFRRFields",
    "PriceAreaAFRRTextFields",
    "PriceAreaDayAhead",
    "PriceAreaDayAheadGraphQL",
    "PriceAreaDayAheadWrite",
    "PriceAreaDayAheadList",
    "PriceAreaDayAheadWriteList",
    "PriceAreaDayAheadFields",
    "PriceAreaDayAheadTextFields",
    "PriceAreaInformation",
    "PriceAreaInformationGraphQL",
    "PriceAreaInformationWrite",
    "PriceAreaInformationList",
    "PriceAreaInformationWriteList",
    "PriceAreaInformationFields",
    "PriceAreaInformationTextFields",
    "PriceProduction",
    "PriceProductionGraphQL",
    "PriceProductionWrite",
    "PriceProductionList",
    "PriceProductionWriteList",
    "PriceProductionFields",
    "PriceProductionTextFields",
    "ShopAttributeMapping",
    "ShopAttributeMappingGraphQL",
    "ShopAttributeMappingWrite",
    "ShopAttributeMappingList",
    "ShopAttributeMappingWriteList",
    "ShopAttributeMappingFields",
    "ShopAttributeMappingTextFields",
    "ShopBasedPartialBidConfiguration",
    "ShopBasedPartialBidConfigurationGraphQL",
    "ShopBasedPartialBidConfigurationWrite",
    "ShopBasedPartialBidConfigurationList",
    "ShopBasedPartialBidConfigurationWriteList",
    "ShopBasedPartialBidConfigurationFields",
    "ShopBasedPartialBidConfigurationTextFields",
    "ShopCase",
    "ShopCaseGraphQL",
    "ShopCaseWrite",
    "ShopCaseList",
    "ShopCaseWriteList",
    "ShopCaseFields",
    "ShopCaseTextFields",
    "ShopCommands",
    "ShopCommandsGraphQL",
    "ShopCommandsWrite",
    "ShopCommandsList",
    "ShopCommandsWriteList",
    "ShopCommandsFields",
    "ShopCommandsTextFields",
    "ShopFile",
    "ShopFileGraphQL",
    "ShopFileWrite",
    "ShopFileList",
    "ShopFileWriteList",
    "ShopFileFields",
    "ShopFileTextFields",
    "ShopModel",
    "ShopModelGraphQL",
    "ShopModelWrite",
    "ShopModelList",
    "ShopModelWriteList",
    "ShopModelFields",
    "ShopModelTextFields",
    "ShopModelWithAssets",
    "ShopModelWithAssetsGraphQL",
    "ShopModelWithAssetsWrite",
    "ShopModelWithAssetsList",
    "ShopModelWithAssetsWriteList",
    "ShopModelWithAssetsFields",
    "ShopModelWithAssetsTextFields",
    "ShopOutputTimeSeriesDefinition",
    "ShopOutputTimeSeriesDefinitionGraphQL",
    "ShopOutputTimeSeriesDefinitionWrite",
    "ShopOutputTimeSeriesDefinitionList",
    "ShopOutputTimeSeriesDefinitionWriteList",
    "ShopOutputTimeSeriesDefinitionFields",
    "ShopOutputTimeSeriesDefinitionTextFields",
    "ShopPenaltyReport",
    "ShopPenaltyReportGraphQL",
    "ShopPenaltyReportWrite",
    "ShopPenaltyReportList",
    "ShopPenaltyReportWriteList",
    "ShopPenaltyReportFields",
    "ShopPenaltyReportTextFields",
    "ShopPreprocessorInput",
    "ShopPreprocessorInputGraphQL",
    "ShopPreprocessorInputWrite",
    "ShopPreprocessorInputList",
    "ShopPreprocessorInputWriteList",
    "ShopPreprocessorInputFields",
    "ShopPreprocessorInputTextFields",
    "ShopPreprocessorOutput",
    "ShopPreprocessorOutputGraphQL",
    "ShopPreprocessorOutputWrite",
    "ShopPreprocessorOutputList",
    "ShopPreprocessorOutputWriteList",
    "ShopPreprocessorOutputFields",
    "ShopPreprocessorOutputTextFields",
    "ShopResult",
    "ShopResultGraphQL",
    "ShopResultWrite",
    "ShopResultList",
    "ShopResultWriteList",
    "ShopResultFields",
    "ShopResultTextFields",
    "ShopScenario",
    "ShopScenarioGraphQL",
    "ShopScenarioWrite",
    "ShopScenarioList",
    "ShopScenarioWriteList",
    "ShopScenarioFields",
    "ShopScenarioTextFields",
    "ShopScenarioSet",
    "ShopScenarioSetGraphQL",
    "ShopScenarioSetWrite",
    "ShopScenarioSetList",
    "ShopScenarioSetWriteList",
    "ShopScenarioSetFields",
    "ShopScenarioSetTextFields",
    "ShopTimeResolution",
    "ShopTimeResolutionGraphQL",
    "ShopTimeResolutionWrite",
    "ShopTimeResolutionList",
    "ShopTimeResolutionWriteList",
    "ShopTimeResolutionFields",
    "ShopTimeResolutionTextFields",
    "ShopTimeSeries",
    "ShopTimeSeriesGraphQL",
    "ShopTimeSeriesWrite",
    "ShopTimeSeriesList",
    "ShopTimeSeriesWriteList",
    "ShopTimeSeriesFields",
    "ShopTimeSeriesTextFields",
    "ShopTriggerInput",
    "ShopTriggerInputGraphQL",
    "ShopTriggerInputWrite",
    "ShopTriggerInputList",
    "ShopTriggerInputWriteList",
    "ShopTriggerInputFields",
    "ShopTriggerInputTextFields",
    "ShopTriggerOutput",
    "ShopTriggerOutputGraphQL",
    "ShopTriggerOutputWrite",
    "ShopTriggerOutputList",
    "ShopTriggerOutputWriteList",
    "ShopTriggerOutputFields",
    "ShopTriggerOutputTextFields",
    "TaskDispatcherInput",
    "TaskDispatcherInputGraphQL",
    "TaskDispatcherInputWrite",
    "TaskDispatcherInputList",
    "TaskDispatcherInputWriteList",
    "TaskDispatcherInputFields",
    "TaskDispatcherInputTextFields",
    "TaskDispatcherOutput",
    "TaskDispatcherOutputGraphQL",
    "TaskDispatcherOutputWrite",
    "TaskDispatcherOutputList",
    "TaskDispatcherOutputWriteList",
    "TaskDispatcherOutputFields",
    "TaskDispatcherOutputTextFields",
    "TotalBidMatrixCalculationInput",
    "TotalBidMatrixCalculationInputGraphQL",
    "TotalBidMatrixCalculationInputWrite",
    "TotalBidMatrixCalculationInputList",
    "TotalBidMatrixCalculationInputWriteList",
    "TotalBidMatrixCalculationInputFields",
    "TotalBidMatrixCalculationInputTextFields",
    "TotalBidMatrixCalculationOutput",
    "TotalBidMatrixCalculationOutputGraphQL",
    "TotalBidMatrixCalculationOutputWrite",
    "TotalBidMatrixCalculationOutputList",
    "TotalBidMatrixCalculationOutputWriteList",
    "TotalBidMatrixCalculationOutputFields",
    "TotalBidMatrixCalculationOutputTextFields",
    "TurbineEfficiencyCurve",
    "TurbineEfficiencyCurveGraphQL",
    "TurbineEfficiencyCurveWrite",
    "TurbineEfficiencyCurveList",
    "TurbineEfficiencyCurveWriteList",
    "TurbineEfficiencyCurveFields",
    "TurbineEfficiencyCurveTextFields",
    "WaterValueBasedPartialBidConfiguration",
    "WaterValueBasedPartialBidConfigurationGraphQL",
    "WaterValueBasedPartialBidConfigurationWrite",
    "WaterValueBasedPartialBidConfigurationList",
    "WaterValueBasedPartialBidConfigurationWriteList",
    "WaterValueBasedPartialBidConfigurationFields",
    "WaterValueBasedPartialBidConfigurationTextFields",
    "WaterValueBasedPartialBidMatrixCalculationInput",
    "WaterValueBasedPartialBidMatrixCalculationInputGraphQL",
    "WaterValueBasedPartialBidMatrixCalculationInputWrite",
    "WaterValueBasedPartialBidMatrixCalculationInputList",
    "WaterValueBasedPartialBidMatrixCalculationInputWriteList",
    "WaterValueBasedPartialBidMatrixCalculationInputFields",
    "WaterValueBasedPartialBidMatrixCalculationInputTextFields",
    "Watercourse",
    "WatercourseGraphQL",
    "WatercourseWrite",
    "WatercourseList",
    "WatercourseWriteList",
    "WatercourseFields",
    "WatercourseTextFields",
]



from __future__ import annotations

import logging
from typing import Literal

import numpy as np
import pandas as pd
from cognite.client import CogniteClient
from cognite.client.data_classes import DatapointsList, DataSet
from cognite.client.utils import ms_to_datetime

from cognite.powerops.client.powerops_client import PowerOpsClient
from cognite.powerops.utils.require import require

DataSetType = Literal["READ", "WRITE", "MONITOR", "PROCESS"]

logger = logging.getLogger(__name__)


def remove_duplicates(lst: list) -> list:
    return list(set(lst))


def _overlapping_keys(dict_: dict, other: dict) -> list[str]:
    return list(set(dict_.keys()).intersection(other.keys()))


def merge_dicts(*args: dict) -> dict:
    merged: dict = {}
    for dict_ in args:
        assert isinstance(dict_, dict), f"Expected dict, got {type(dict_)}!"
        overlap = _overlapping_keys(dict_, merged)
        if overlap:
            raise Exception(f"Key collision on '{overlap}' when merging dictionaries!")
        merged = {**merged, **dict_}
    return merged


def _retrieve_range(client: CogniteClient, external_ids: list[str], start: int, end: int) -> pd.DataFrame:
    # TODO: Upgrade cognite-sdk to v5 (or later), and see how much of the code we can replace with direct SDK calls
    # - client.time_series.data.retrieve_dataframe(…, uniform_index=True) should give us almost what we want,
    # but maybe we need to be careful with cases where there is more than 1 hour between values
    # (I do not remember if this is an issue only for some aggregates like average, or for all).
    # And maybe we need to parametrise the "minimum resolution"
    # (seems to assume 1 hour, but we should support sub-hourly resolution)
    # Retrieve raw datapoints
    external_ids = remove_duplicates(external_ids)
    if not external_ids:
        return pd.DataFrame()

    start_dt = ms_to_datetime(start).replace(tzinfo=None)  # UTC implied
    end_dt = ms_to_datetime(end).replace(tzinfo=None)  # UTC implied
    logger.debug(f"Retrieving {external_ids} between '{start_dt}' and '{end_dt}'")
    df_range = client.time_series.data.retrieve(
        external_id=external_ids, start=start, end=end, ignore_unknown_ids=True
    ).to_pandas()

    # Retrieve latest datapoints before start
    df_latest = client.time_series.data.retrieve_latest(
        external_id=external_ids, before=start, ignore_unknown_ids=True
    ).to_pandas()

    # Make sure we have a start timestamp in range
    if df_range.empty:
        df_range = pd.DataFrame(
            columns=df_latest.columns,
            index=pd.DatetimeIndex(data=np.array([int(start)], dtype="datetime64[ms]")),
            dtype=float,
        )

    # Add the latest datapoints to the DataFrame
    df_raw = df_range.combine_first(df_latest)

    # Must retrieve time series metadata to correctly resample and aggregate datapoints
    time_series = client.time_series.retrieve_multiple(external_ids=external_ids, ignore_unknown_ids=True)
    step_columns = [ts.external_id for ts in time_series if ts.is_step]
    linear_columns = [ts.external_id for ts in time_series if not ts.is_step]
    logger.debug(f"time_series.is_step: True [{len(step_columns)}] False [{len(linear_columns)}]")

    # Step interpolation of time series with .is_step=False
    # NOTE: do not need to upsample when forward filling
    # TODO: note 2x ffill()
    df_step = df_raw[step_columns].ffill().resample("1h").ffill()

    # Linear interpolation of time series with .is_step=False
    # TODO: must upsample before downsampling?
    # TODO: confirm operations
    intermediate_df = df_raw[linear_columns].resample("1min").interpolate()
    df_linear = intermediate_df.resample("1h").interpolate()

    # Merge the step interpolated and linearly interpolated DataFrames
    df_combined = df_step.combine_first(df_linear)

    # Only return datapoints within the range
    df_filtered = df_combined[start_dt:end_dt]  # type: ignore[misc]

    return df_filtered


def retrieve_range(client: CogniteClient, external_ids: list[str], start: int, end: int) -> dict[str, pd.Series]:
    retrieved_range_df = _retrieve_range(client=client, external_ids=external_ids, start=start, end=end)
    return {col: retrieved_range_df[col].dropna() for col in retrieved_range_df.columns}


# TODO: refactor
# TODO: naming: time_series vs. datapoints
def retrieve_latest(client: CogniteClient, external_ids: list[str | None], before: int) -> dict[str, pd.Series]:
    if not external_ids:
        return {}
    external_ids = remove_duplicates(external_ids)
    logger.debug(f"Retrieving {external_ids} before '{ms_to_datetime(before)}'")
    time_series: DatapointsList = client.time_series.data.retrieve_latest(
        external_id=external_ids, before=before, ignore_unknown_ids=True
    )

    # For (Cog)Datapoints in (Cog)DatapointsList
    for datapoints in time_series:
        if len(datapoints) > 0:  # TODO: what to do about ts with no datapoints?
            datapoints.timestamp[0] = before

    res = {
        require(datapoints.external_id): datapoints.to_pandas().iloc[:, 0]  # iloc to convert DataFrame to Series
        for datapoints in time_series
        if len(datapoints) > 0
    }
    if missing := set(external_ids).difference(res):
        logger.warning(f"Missing: {', '.join(map(str, missing))}")
    return res


def retrieve_time_series_datapoints(  # type: ignore[no-untyped-def]
    client: CogniteClient, mappings, start: int, end: int
) -> dict[str, pd.Series]:
    time_series_start = retrieve_latest(
        client=client,
        external_ids=[mapping.cdf_time_series for mapping in mappings if mapping.retrieve == "START"],
        before=start,
    )
    time_series_end = retrieve_latest(
        client=client,
        external_ids=[mapping.cdf_time_series for mapping in mappings if mapping.retrieve == "END"],
        before=end,
    )
    time_series_range = retrieve_range(
        client=client,
        external_ids=[mapping.cdf_time_series for mapping in mappings if mapping.retrieve == "RANGE"],
        start=start,
        end=end,
    )
    _time_series_none = [
        ".".join(filter(None, [mapping.shop_object_type, mapping.shop_object_name, mapping.shop_attribute_name]))
        for mapping in mappings
        if not mapping.retrieve
    ]
    logger.debug(f"Not retrieving datapoints for {_time_series_none}")

    return merge_dicts(time_series_start, time_series_end, time_series_range)


def get_data_set_from_config(client: PowerOpsClient, data_set_type: DataSetType = "READ") -> DataSet:
    """
    Get the latest dataset configuration for a given data_set_type.

    Args:
        client (PowerOpsClient): The PowerOps client instance.
        data_set_type (DataSetType): The type of dataset to retrieve. Valid options
                                     are "READ", "WRITE", "MONITOR", "PROCESS".

    Returns:
        DataSet: The dataset object corresponding to the specified data_set_type.

    Raises:
        ValueError: If no dataset configuration is found, if the specified data_set_type is invalid, or
                    if the dataset with the specified external_id is not found.
    """
    dataset_configs = client.powermodel.day_ahead_configuration.data_set_configuration.list(limit=-1)
    if not dataset_configs:
        raise ValueError("No dataset configuration found.")

    # Pick the config with the latest last_updated_time
    dataset_config = max(
        dataset_configs,
        key=lambda cfg: getattr(cfg.data_record, "last_updated_time", 0),
    )

    if data_set_type == "READ":
        external_id = dataset_config.read_data_set
    elif data_set_type == "WRITE":
        external_id = dataset_config.write_data_set
    elif data_set_type == "MONITOR":
        external_id = dataset_config.monitor_data_set
    elif data_set_type == "PROCESS":
        external_id = dataset_config.process_data_set
    else:
        raise ValueError(f"Unknown data_set_type: {data_set_type}")

    if not external_id:
        raise ValueError(f"No external_id found for data_set_type: {data_set_type}")

    dataset = client.cdf.data_sets.retrieve(external_id=external_id)

    if dataset is None:
        raise ValueError(f"Dataset with external_id '{external_id}' not found.")

    return dataset



from __future__ import annotations

from collections.abc import Sequence
from typing import Any, ClassVar, Literal, Optional, Union

from cognite.client import data_modeling as dm, CogniteClient
from pydantic import field_validator, model_validator, ValidationInfo

from cognite.powerops.client._generated.config import global_config
from cognite.powerops.client._generated.data_classes._core import (
    DEFAULT_INSTANCE_SPACE,
    DEFAULT_QUERY_LIMIT,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelation,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
    T_DomainModelList,
    as_node_id,
    as_read_args,
    as_write_args,
    is_tuple_id,
    as_instance_dict_id,
    parse_single_connection,
    QueryCore,
    NodeQueryCore,
    StringFilter,
    ViewPropertyId,
)


__all__ = [
    "GeneratorEfficiencyCurve",
    "GeneratorEfficiencyCurveWrite",
    "GeneratorEfficiencyCurveList",
    "GeneratorEfficiencyCurveWriteList",
    "GeneratorEfficiencyCurveFields",
    "GeneratorEfficiencyCurveGraphQL",
]


GeneratorEfficiencyCurveTextFields = Literal["external_id", ]
GeneratorEfficiencyCurveFields = Literal["external_id", "power", "efficiency"]

_GENERATOREFFICIENCYCURVE_PROPERTIES_BY_FIELD = {
    "external_id": "externalId",
    "power": "power",
    "efficiency": "efficiency",
}


class GeneratorEfficiencyCurveGraphQL(GraphQLCore):
    """This represents the reading version of generator efficiency curve, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the generator efficiency curve.
        data_record: The data record of the generator efficiency curve node.
        power: The generator power values
        efficiency: The generator efficiency values
    """

    view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "GeneratorEfficiencyCurve", "1")
    power: Optional[list[float]] = None
    efficiency: Optional[list[float]] = None

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values



    def as_read(self) -> GeneratorEfficiencyCurve:
        """Convert this GraphQL format of generator efficiency curve to the reading format."""
        return GeneratorEfficiencyCurve.model_validate(as_read_args(self))

    def as_write(self) -> GeneratorEfficiencyCurveWrite:
        """Convert this GraphQL format of generator efficiency curve to the writing format."""
        return GeneratorEfficiencyCurveWrite.model_validate(as_write_args(self))


class GeneratorEfficiencyCurve(DomainModel):
    """This represents the reading version of generator efficiency curve.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the generator efficiency curve.
        data_record: The data record of the generator efficiency curve node.
        power: The generator power values
        efficiency: The generator efficiency values
    """

    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "GeneratorEfficiencyCurve", "1")

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("power_ops_types", "GeneratorEfficiencyCurve")
    power: list[float]
    efficiency: list[float]


    def as_write(self) -> GeneratorEfficiencyCurveWrite:
        """Convert this read version of generator efficiency curve to the writing version."""
        return GeneratorEfficiencyCurveWrite.model_validate(as_write_args(self))



class GeneratorEfficiencyCurveWrite(DomainModelWrite):
    """This represents the writing version of generator efficiency curve.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the generator efficiency curve.
        data_record: The data record of the generator efficiency curve node.
        power: The generator power values
        efficiency: The generator efficiency values
    """
    _container_fields: ClassVar[tuple[str, ...]] = ("efficiency", "power",)

    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "GeneratorEfficiencyCurve", "1")

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, dm.NodeId, tuple[str, str], None] = dm.DirectRelationReference("power_ops_types", "GeneratorEfficiencyCurve")
    power: list[float]
    efficiency: list[float]



class GeneratorEfficiencyCurveList(DomainModelList[GeneratorEfficiencyCurve]):
    """List of generator efficiency curves in the read version."""

    _INSTANCE = GeneratorEfficiencyCurve
    def as_write(self) -> GeneratorEfficiencyCurveWriteList:
        """Convert these read versions of generator efficiency curve to the writing versions."""
        return GeneratorEfficiencyCurveWriteList([node.as_write() for node in self.data])



class GeneratorEfficiencyCurveWriteList(DomainModelWriteList[GeneratorEfficiencyCurveWrite]):
    """List of generator efficiency curves in the writing version."""

    _INSTANCE = GeneratorEfficiencyCurveWrite


def _create_generator_efficiency_curve_filter(
    view_id: dm.ViewId,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters: list[dm.Filter] = []
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None


class _GeneratorEfficiencyCurveQuery(NodeQueryCore[T_DomainModelList, GeneratorEfficiencyCurveList]):
    _view_id = GeneratorEfficiencyCurve._view_id
    _result_cls = GeneratorEfficiencyCurve
    _result_list_cls_end = GeneratorEfficiencyCurveList

    def __init__(
        self,
        created_types: set[type],
        creation_path: list[QueryCore],
        client: CogniteClient,
        result_list_cls: type[T_DomainModelList],
        expression: dm.query.NodeOrEdgeResultSetExpression | None = None,
        connection_name: str | None = None,
        connection_property: ViewPropertyId | None = None,
        connection_type: Literal["reverse-list"] | None = None,
        reverse_expression: dm.query.NodeOrEdgeResultSetExpression | None = None,
    ):

        super().__init__(
            created_types,
            creation_path,
            client,
            result_list_cls,
            expression,
            dm.filters.HasData(views=[self._view_id]),
            connection_name,
            connection_property,
            connection_type,
            reverse_expression,
        )

        self.space = StringFilter(self, ["node", "space"])
        self.external_id = StringFilter(self, ["node", "externalId"])
        self._filter_classes.extend([
            self.space,
            self.external_id,
        ])

    def list_generator_efficiency_curve(self, limit: int = DEFAULT_QUERY_LIMIT) -> GeneratorEfficiencyCurveList:
        return self._list(limit=limit)


class GeneratorEfficiencyCurveQuery(_GeneratorEfficiencyCurveQuery[GeneratorEfficiencyCurveList]):
    def __init__(self, client: CogniteClient):
        super().__init__(set(), [], client, GeneratorEfficiencyCurveList)



from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, ClassVar, Literal, Optional, Union

from cognite.client import data_modeling as dm, CogniteClient
from cognite.client.data_classes import (
    TimeSeries as CogniteTimeSeries,
    TimeSeriesWrite as CogniteTimeSeriesWrite,
)
from pydantic import Field
from pydantic import field_validator, model_validator, ValidationInfo

from cognite.powerops.client._generated.config import global_config
from cognite.powerops.client._generated.data_classes._core import (
    DEFAULT_INSTANCE_SPACE,
    DEFAULT_QUERY_LIMIT,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelation,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
    FileMetadata,
    FileMetadataWrite,
    FileMetadataGraphQL,
    TimeSeries,
    TimeSeriesWrite,
    TimeSeriesGraphQL,
    TimeSeriesReferenceAPI,
    T_DomainModelList,
    as_node_id,
    as_read_args,
    as_write_args,
    is_tuple_id,
    as_instance_dict_id,
    parse_single_connection,
    QueryCore,
    NodeQueryCore,
    StringFilter,
    ViewPropertyId,
    DirectRelationFilter,
    FloatFilter,
    IntFilter,
)
from cognite.powerops.client._generated.data_classes._power_asset import PowerAsset, PowerAssetWrite
if TYPE_CHECKING:
    from cognite.powerops.client._generated.data_classes._generator_efficiency_curve import GeneratorEfficiencyCurve, GeneratorEfficiencyCurveList, GeneratorEfficiencyCurveGraphQL, GeneratorEfficiencyCurveWrite, GeneratorEfficiencyCurveWriteList
    from cognite.powerops.client._generated.data_classes._turbine_efficiency_curve import TurbineEfficiencyCurve, TurbineEfficiencyCurveList, TurbineEfficiencyCurveGraphQL, TurbineEfficiencyCurveWrite, TurbineEfficiencyCurveWriteList


__all__ = [
    "Generator",
    "GeneratorWrite",
    "GeneratorList",
    "GeneratorWriteList",
    "GeneratorFields",
    "GeneratorTextFields",
    "GeneratorGraphQL",
]


GeneratorTextFields = Literal["external_id", "name", "display_name", "asset_type", "start_stop_cost_time_series", "availability_time_series"]
GeneratorFields = Literal["external_id", "name", "display_name", "ordering", "asset_type", "production_min", "penstock_number", "start_stop_cost", "start_stop_cost_time_series", "availability_time_series", "production_max"]

_GENERATOR_PROPERTIES_BY_FIELD = {
    "external_id": "externalId",
    "name": "name",
    "display_name": "displayName",
    "ordering": "ordering",
    "asset_type": "assetType",
    "production_min": "productionMin",
    "penstock_number": "penstockNumber",
    "start_stop_cost": "startStopCost",
    "start_stop_cost_time_series": "startStopCostTimeSeries",
    "availability_time_series": "availabilityTimeSeries",
    "production_max": "productionMax",
}


class GeneratorGraphQL(GraphQLCore):
    """This represents the reading version of generator, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the generator.
        data_record: The data record of the generator node.
        name: Name for the PowerAsset
        display_name: Display name for the PowerAsset.
        ordering: The ordering of the asset
        asset_type: The type of the asset
        production_min: The production min field.
        penstock_number: The penstock number field.
        start_stop_cost: The start stop cost field.
        start_stop_cost_time_series: The start stop cost time series field.
        availability_time_series: The availability time series field.
        generator_efficiency_curve: The generator efficiency curve field.
        production_max: The production max field.
        turbine_efficiency_curves: TODO description
    """

    view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "Generator", "1")
    name: Optional[str] = None
    display_name: Optional[str] = Field(None, alias="displayName")
    ordering: Optional[int] = None
    asset_type: Optional[str] = Field(None, alias="assetType")
    production_min: Optional[float] = Field(None, alias="productionMin")
    penstock_number: Optional[int] = Field(None, alias="penstockNumber")
    start_stop_cost: Optional[float] = Field(None, alias="startStopCost")
    start_stop_cost_time_series: Optional[TimeSeriesGraphQL] = Field(None, alias="startStopCostTimeSeries")
    availability_time_series: Optional[TimeSeriesGraphQL] = Field(None, alias="availabilityTimeSeries")
    generator_efficiency_curve: Optional[GeneratorEfficiencyCurveGraphQL] = Field(default=None, repr=False, alias="generatorEfficiencyCurve")
    production_max: Optional[float] = Field(None, alias="productionMax")
    turbine_efficiency_curves: Optional[list[TurbineEfficiencyCurveGraphQL]] = Field(default=None, repr=False, alias="turbineEfficiencyCurves")

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values


    @field_validator("generator_efficiency_curve", "turbine_efficiency_curves", mode="before")
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    def as_read(self) -> Generator:
        """Convert this GraphQL format of generator to the reading format."""
        return Generator.model_validate(as_read_args(self))

    def as_write(self) -> GeneratorWrite:
        """Convert this GraphQL format of generator to the writing format."""
        return GeneratorWrite.model_validate(as_write_args(self))


class Generator(PowerAsset):
    """This represents the reading version of generator.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the generator.
        data_record: The data record of the generator node.
        name: Name for the PowerAsset
        display_name: Display name for the PowerAsset.
        ordering: The ordering of the asset
        asset_type: The type of the asset
        production_min: The production min field.
        penstock_number: The penstock number field.
        start_stop_cost: The start stop cost field.
        start_stop_cost_time_series: The start stop cost time series field.
        availability_time_series: The availability time series field.
        generator_efficiency_curve: The generator efficiency curve field.
        production_max: The production max field.
        turbine_efficiency_curves: TODO description
    """

    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "Generator", "1")

    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("power_ops_types", "Generator")
    production_min: float = Field(alias="productionMin")
    penstock_number: int = Field(alias="penstockNumber")
    start_stop_cost: float = Field(alias="startStopCost")
    start_stop_cost_time_series: Union[TimeSeries, str, None] = Field(None, alias="startStopCostTimeSeries")
    availability_time_series: Union[TimeSeries, str, None] = Field(None, alias="availabilityTimeSeries")
    generator_efficiency_curve: Union[GeneratorEfficiencyCurve, str, dm.NodeId, None] = Field(default=None, repr=False, alias="generatorEfficiencyCurve")
    production_max: Optional[float] = Field(None, alias="productionMax")
    turbine_efficiency_curves: Optional[list[Union[TurbineEfficiencyCurve, str, dm.NodeId]]] = Field(default=None, repr=False, alias="turbineEfficiencyCurves")
    @field_validator("generator_efficiency_curve", mode="before")
    @classmethod
    def parse_single(cls, value: Any, info: ValidationInfo) -> Any:
        return parse_single_connection(value, info.field_name)

    @field_validator("turbine_efficiency_curves", mode="before")
    @classmethod
    def parse_list(cls, value: Any, info: ValidationInfo) -> Any:
        if value is None:
            return None
        return [parse_single_connection(item, info.field_name) for item in value]

    def as_write(self) -> GeneratorWrite:
        """Convert this read version of generator to the writing version."""
        return GeneratorWrite.model_validate(as_write_args(self))



class GeneratorWrite(PowerAssetWrite):
    """This represents the writing version of generator.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the generator.
        data_record: The data record of the generator node.
        name: Name for the PowerAsset
        display_name: Display name for the PowerAsset.
        ordering: The ordering of the asset
        asset_type: The type of the asset
        production_min: The production min field.
        penstock_number: The penstock number field.
        start_stop_cost: The start stop cost field.
        start_stop_cost_time_series: The start stop cost time series field.
        availability_time_series: The availability time series field.
        generator_efficiency_curve: The generator efficiency curve field.
        production_max: The production max field.
        turbine_efficiency_curves: TODO description
    """
    _container_fields: ClassVar[tuple[str, ...]] = ("asset_type", "availability_time_series", "display_name", "generator_efficiency_curve", "name", "ordering", "penstock_number", "production_max", "production_min", "start_stop_cost", "start_stop_cost_time_series",)
    _outwards_edges: ClassVar[tuple[tuple[str, dm.DirectRelationReference], ...]] = (("turbine_efficiency_curves", dm.DirectRelationReference("power_ops_types", "isSubAssetOf")),)
    _direct_relations: ClassVar[tuple[str, ...]] = ("generator_efficiency_curve",)

    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "Generator", "1")

    node_type: Union[dm.DirectRelationReference, dm.NodeId, tuple[str, str], None] = dm.DirectRelationReference("power_ops_types", "Generator")
    production_min: float = Field(alias="productionMin")
    penstock_number: int = Field(alias="penstockNumber")
    start_stop_cost: float = Field(alias="startStopCost")
    start_stop_cost_time_series: Union[TimeSeriesWrite, str, None] = Field(None, alias="startStopCostTimeSeries")
    availability_time_series: Union[TimeSeriesWrite, str, None] = Field(None, alias="availabilityTimeSeries")
    generator_efficiency_curve: Union[GeneratorEfficiencyCurveWrite, str, dm.NodeId, None] = Field(default=None, repr=False, alias="generatorEfficiencyCurve")
    production_max: Optional[float] = Field(None, alias="productionMax")
    turbine_efficiency_curves: Optional[list[Union[TurbineEfficiencyCurveWrite, str, dm.NodeId]]] = Field(default=None, repr=False, alias="turbineEfficiencyCurves")

    @field_validator("generator_efficiency_curve", "turbine_efficiency_curves", mode="before")
    def as_node_id(cls, value: Any) -> Any:
        if isinstance(value, dm.DirectRelationReference):
            return dm.NodeId(value.space, value.external_id)
        elif isinstance(value, tuple) and len(value) == 2 and all(isinstance(item, str) for item in value):
            return dm.NodeId(value[0], value[1])
        elif isinstance(value, list):
            return [cls.as_node_id(item) for item in value]
        return value


class GeneratorList(DomainModelList[Generator]):
    """List of generators in the read version."""

    _INSTANCE = Generator
    def as_write(self) -> GeneratorWriteList:
        """Convert these read versions of generator to the writing versions."""
        return GeneratorWriteList([node.as_write() for node in self.data])


    @property
    def generator_efficiency_curve(self) -> GeneratorEfficiencyCurveList:
        from ._generator_efficiency_curve import GeneratorEfficiencyCurve, GeneratorEfficiencyCurveList
        return GeneratorEfficiencyCurveList([item.generator_efficiency_curve for item in self.data if isinstance(item.generator_efficiency_curve, GeneratorEfficiencyCurve)])
    @property
    def turbine_efficiency_curves(self) -> TurbineEfficiencyCurveList:
        from ._turbine_efficiency_curve import TurbineEfficiencyCurve, TurbineEfficiencyCurveList
        return TurbineEfficiencyCurveList([item for items in self.data for item in items.turbine_efficiency_curves or [] if isinstance(item, TurbineEfficiencyCurve)])


class GeneratorWriteList(DomainModelWriteList[GeneratorWrite]):
    """List of generators in the writing version."""

    _INSTANCE = GeneratorWrite
    @property
    def generator_efficiency_curve(self) -> GeneratorEfficiencyCurveWriteList:
        from ._generator_efficiency_curve import GeneratorEfficiencyCurveWrite, GeneratorEfficiencyCurveWriteList
        return GeneratorEfficiencyCurveWriteList([item.generator_efficiency_curve for item in self.data if isinstance(item.generator_efficiency_curve, GeneratorEfficiencyCurveWrite)])
    @property
    def turbine_efficiency_curves(self) -> TurbineEfficiencyCurveWriteList:
        from ._turbine_efficiency_curve import TurbineEfficiencyCurveWrite, TurbineEfficiencyCurveWriteList
        return TurbineEfficiencyCurveWriteList([item for items in self.data for item in items.turbine_efficiency_curves or [] if isinstance(item, TurbineEfficiencyCurveWrite)])



def _create_generator_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    display_name: str | list[str] | None = None,
    display_name_prefix: str | None = None,
    min_ordering: int | None = None,
    max_ordering: int | None = None,
    asset_type: str | list[str] | None = None,
    asset_type_prefix: str | None = None,
    min_production_min: float | None = None,
    max_production_min: float | None = None,
    min_penstock_number: int | None = None,
    max_penstock_number: int | None = None,
    min_start_stop_cost: float | None = None,
    max_start_stop_cost: float | None = None,
    generator_efficiency_curve: str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference | Sequence[str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference] | None = None,
    min_production_max: float | None = None,
    max_production_max: float | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters: list[dm.Filter] = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if isinstance(display_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("displayName"), value=display_name))
    if display_name and isinstance(display_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("displayName"), values=display_name))
    if display_name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("displayName"), value=display_name_prefix))
    if min_ordering is not None or max_ordering is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("ordering"), gte=min_ordering, lte=max_ordering))
    if isinstance(asset_type, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("assetType"), value=asset_type))
    if asset_type and isinstance(asset_type, list):
        filters.append(dm.filters.In(view_id.as_property_ref("assetType"), values=asset_type))
    if asset_type_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("assetType"), value=asset_type_prefix))
    if min_production_min is not None or max_production_min is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("productionMin"), gte=min_production_min, lte=max_production_min))
    if min_penstock_number is not None or max_penstock_number is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("penstockNumber"), gte=min_penstock_number, lte=max_penstock_number))
    if min_start_stop_cost is not None or max_start_stop_cost is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("startStopCost"), gte=min_start_stop_cost, lte=max_start_stop_cost))
    if isinstance(generator_efficiency_curve, str | dm.NodeId | dm.DirectRelationReference) or is_tuple_id(generator_efficiency_curve):
        filters.append(dm.filters.Equals(view_id.as_property_ref("generatorEfficiencyCurve"), value=as_instance_dict_id(generator_efficiency_curve)))
    if generator_efficiency_curve and isinstance(generator_efficiency_curve, Sequence) and not isinstance(generator_efficiency_curve, str) and not is_tuple_id(generator_efficiency_curve):
        filters.append(dm.filters.In(view_id.as_property_ref("generatorEfficiencyCurve"), values=[as_instance_dict_id(item) for item in generator_efficiency_curve]))
    if min_production_max is not None or max_production_max is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("productionMax"), gte=min_production_max, lte=max_production_max))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None


class _GeneratorQuery(NodeQueryCore[T_DomainModelList, GeneratorList]):
    _view_id = Generator._view_id
    _result_cls = Generator
    _result_list_cls_end = GeneratorList

    def __init__(
        self,
        created_types: set[type],
        creation_path: list[QueryCore],
        client: CogniteClient,
        result_list_cls: type[T_DomainModelList],
        expression: dm.query.NodeOrEdgeResultSetExpression | None = None,
        connection_name: str | None = None,
        connection_property: ViewPropertyId | None = None,
        connection_type: Literal["reverse-list"] | None = None,
        reverse_expression: dm.query.NodeOrEdgeResultSetExpression | None = None,
    ):
        from ._generator_efficiency_curve import _GeneratorEfficiencyCurveQuery
        from ._turbine_efficiency_curve import _TurbineEfficiencyCurveQuery

        super().__init__(
            created_types,
            creation_path,
            client,
            result_list_cls,
            expression,
            dm.filters.HasData(views=[self._view_id]),
            connection_name,
            connection_property,
            connection_type,
            reverse_expression,
        )

        if _GeneratorEfficiencyCurveQuery not in created_types and len(creation_path) + 1 < global_config.max_select_depth:
            self.generator_efficiency_curve = _GeneratorEfficiencyCurveQuery(
                created_types.copy(),
                self._creation_path,
                client,
                result_list_cls,
                dm.query.NodeResultSetExpression(
                    through=self._view_id.as_property_ref("generatorEfficiencyCurve"),
                    direction="outwards",
                ),
                connection_name="generator_efficiency_curve",
                connection_property=ViewPropertyId(self._view_id, "generatorEfficiencyCurve"),
            )

        if _TurbineEfficiencyCurveQuery not in created_types and len(creation_path) + 1 < global_config.max_select_depth:
            self.turbine_efficiency_curves = _TurbineEfficiencyCurveQuery(
                created_types.copy(),
                self._creation_path,
                client,
                result_list_cls,
                dm.query.EdgeResultSetExpression(
                    direction="outwards",
                    chain_to="destination",
                ),
                connection_name="turbine_efficiency_curves",
                connection_property=ViewPropertyId(self._view_id, "turbineEfficiencyCurves"),
            )

        self.space = StringFilter(self, ["node", "space"])
        self.external_id = StringFilter(self, ["node", "externalId"])
        self.name = StringFilter(self, self._view_id.as_property_ref("name"))
        self.display_name = StringFilter(self, self._view_id.as_property_ref("displayName"))
        self.ordering = IntFilter(self, self._view_id.as_property_ref("ordering"))
        self.asset_type = StringFilter(self, self._view_id.as_property_ref("assetType"))
        self.production_min = FloatFilter(self, self._view_id.as_property_ref("productionMin"))
        self.penstock_number = IntFilter(self, self._view_id.as_property_ref("penstockNumber"))
        self.start_stop_cost = FloatFilter(self, self._view_id.as_property_ref("startStopCost"))
        self.generator_efficiency_curve_filter = DirectRelationFilter(self, self._view_id.as_property_ref("generatorEfficiencyCurve"))
        self.production_max = FloatFilter(self, self._view_id.as_property_ref("productionMax"))
        self._filter_classes.extend([
            self.space,
            self.external_id,
            self.name,
            self.display_name,
            self.ordering,
            self.asset_type,
            self.production_min,
            self.penstock_number,
            self.start_stop_cost,
            self.generator_efficiency_curve_filter,
            self.production_max,
        ])
        self.start_stop_cost_time_series = TimeSeriesReferenceAPI(client,  lambda limit: [
            item.start_stop_cost_time_series if isinstance(item.start_stop_cost_time_series, str) else item.start_stop_cost_time_series.external_id #type: ignore[misc]
            for item in self._list(limit=limit)
            if item.start_stop_cost_time_series is not None and
               (isinstance(item.start_stop_cost_time_series, str) or item.start_stop_cost_time_series.external_id is not None)
        ])
        self.availability_time_series = TimeSeriesReferenceAPI(client,  lambda limit: [
            item.availability_time_series if isinstance(item.availability_time_series, str) else item.availability_time_series.external_id #type: ignore[misc]
            for item in self._list(limit=limit)
            if item.availability_time_series is not None and
               (isinstance(item.availability_time_series, str) or item.availability_time_series.external_id is not None)
        ])

    def list_generator(self, limit: int = DEFAULT_QUERY_LIMIT) -> GeneratorList:
        return self._list(limit=limit)


class GeneratorQuery(_GeneratorQuery[GeneratorList]):
    def __init__(self, client: CogniteClient):
        super().__init__(set(), [], client, GeneratorList)



from __future__ import annotations

import logging
from pathlib import Path

from cognite.client import CogniteClient
from cognite.client.data_classes.data_modeling import EdgeId, NodeId, ViewId
from cognite.client.data_classes.filters import In

from cognite.powerops.utils.serialization import load_yaml

logger = logging.getLogger("resync")


class ResyncPurge:
    """Class for purging resync data from CDF."""

    toolkit_modules: list[Path]
    models_space: str
    data_model_version: str
    client: CogniteClient
    exclude_edges: list[str]
    exclude_nodes: list[str]
    dry_run: bool = False
    verbose: bool = False

    def __init__(
        self,
        toolkit_directory: list[Path],
        dry_run: bool,
        verbose: bool,
        client: CogniteClient,
        models_space: str,
        type_space: str,
        data_model_version: str,
        exclude_edges: list[str],
        exclude_nodes: list[str],
    ):
        self.toolkit_directory = toolkit_directory
        self.dry_run = dry_run
        self.client = client
        self.data_model_version = data_model_version
        self.models_space = models_space
        self.type_space = type_space
        self.verbose = verbose
        self.exclude_edges = exclude_edges
        self.exclude_nodes = exclude_nodes

    @classmethod
    def from_yaml(
        cls,
        configuration_path: Path,
        cdf_client: CogniteClient,
        dry_run: bool = False,
        verbose: bool = False,
    ) -> ResyncPurge:
        """Creates a ResyncImporter object from a resync configuration file.

        Args:
            configuration_path: Path to the resync configuration file.
            data_model_classes: A dictionary of all data model classes to be used for the resync configuration.

        Returns:
            A ResyncImporter object.
        """
        configuration = load_yaml(configuration_path, expected_return_type="dict")

        temp_toolkit_directory = configuration.get("toolkit_modules")
        if temp_toolkit_directory:
            if isinstance(temp_toolkit_directory, str):
                toolkit_directory = [Path(temp_toolkit_directory)]
            elif isinstance(temp_toolkit_directory, list):
                toolkit_directory = [Path(directory) for directory in temp_toolkit_directory]
            else:
                raise ValueError("toolkit_directory must be a string or a list of strings")
        else:
            raise ValueError("toolkit_directory is required in the configuration file")

        return cls(
            toolkit_directory=toolkit_directory,
            dry_run=dry_run,
            verbose=verbose,
            client=cdf_client,
            data_model_version=configuration["data_model_version"],
            models_space=configuration["models_space"],
            type_space=configuration["type_space"],
            exclude_nodes=configuration.get("exclude_nodes", []),
            exclude_edges=configuration.get("exclude_edges", []),
        )

    def purge(self) -> None:
        """Purges resync data from CDF."""

        node_ids, edge_ids = self.get_toolkit_external_ids()
        views_with_nodes_to_delete = self.get_nodes_to_delete(node_ids)

        logger.info("NODES to be deleted:")
        all_nodes_to_delete: list[NodeId] = []
        for view, nodes in views_with_nodes_to_delete.items():
            logger.info(f"- {view}: {len(nodes)}")
            if self.verbose:
                logger.info(f"   * External ids: {nodes}")
            all_nodes_to_delete.extend(nodes)

        logger.info("EDGES to be deleted:")
        all_edges_to_delete = []
        edges_to_delete = self.get_edges_to_delete(edge_ids)
        for edge_type, edges in edges_to_delete.items():
            logger.info(f"- {edge_type}: {len(edges)}")
            if self.verbose:
                logger.info(f"   * External ids of edge type {edge_type}: {edges}")
            all_edges_to_delete.extend(edges)

        logger.info("Purge Summary:")
        logger.info(f"- Found {len(all_nodes_to_delete)} nodes in total to delete.")
        logger.info(f"- Found {len(all_edges_to_delete)} edges in total to delete.")

        if self.dry_run:
            logger.info("Dry run mode enabled. Exiting without deleting any data.")
            return

        self.client.data_modeling.instances.delete(nodes=all_nodes_to_delete, edges=all_edges_to_delete)
        logger.info(f"Deleted {len(all_nodes_to_delete)} nodes and {len(all_edges_to_delete)} edges.")

    def get_toolkit_external_ids(self) -> tuple[dict[str, list[tuple[str, str]]], dict[str, list[tuple[str, str]]]]:
        """Get all node and external IDs from the toolkit files."""

        node_external_ids: dict[str, list[tuple[str, str]]] = {}
        edge_external_ids: dict[str, list[tuple[str, str]]] = {}
        for toolkit_directory in self.toolkit_directory:
            yaml_files = list(toolkit_directory.rglob("*.yaml")) + list(toolkit_directory.rglob("*.yml"))
            for file in yaml_files:
                instances = load_yaml(file, expected_return_type="list")
                for instance in instances:
                    if instance_type := instance.get("instanceType"):
                        external_id: tuple[str, str] = (instance["space"], instance["externalId"])
                        if instance_type == "node":
                            node_type = instance.get("type")
                            if node_type:
                                node_type_xid = node_type["externalId"]
                                if node_type_xid in node_external_ids:
                                    temp_list = node_external_ids[node_type_xid]
                                    temp_list.append(external_id)
                                    node_external_ids[node_type_xid] = temp_list
                                else:
                                    node_external_ids[node_type_xid] = [external_id]
                        elif instance_type == "edge":
                            edge_type = instance.get("type")
                            if edge_type:
                                edge_type_xid = edge_type["externalId"]
                                if edge_type_xid in edge_external_ids:
                                    temp_list = edge_external_ids[edge_type_xid]
                                    temp_list.append(external_id)
                                    edge_external_ids[edge_type_xid] = temp_list
                                else:
                                    edge_external_ids[edge_type_xid] = [external_id]

        return node_external_ids, edge_external_ids

    def get_nodes_to_delete(self, node_ids: dict[str, list[tuple[str, str]]]) -> dict[str, list[NodeId]]:
        """Gets all the nodes that exist that are not in the toolkit files."""

        views_with_nodes_to_delete: dict[str, list[NodeId]] = {}
        for view_xid in node_ids.keys():
            if view_xid not in self.exclude_nodes:
                view_id = ViewId(self.models_space, view_xid, self.data_model_version)
                existing_nodes = self.client.data_modeling.instances.list(
                    instance_type="node", limit=None, sources=view_id
                )
                toolkit_nodes = node_ids.get(view_xid, [])
                for node in existing_nodes:
                    temp_node = (node.space, node.external_id)
                    if temp_node not in toolkit_nodes:
                        if view_xid in views_with_nodes_to_delete:
                            temp_list = views_with_nodes_to_delete[view_xid]
                            temp_list.append(node.as_id())
                            views_with_nodes_to_delete[view_xid] = temp_list
                        else:
                            views_with_nodes_to_delete[view_xid] = [node.as_id()]

        return views_with_nodes_to_delete

    def get_edges_to_delete(self, edge_ids: dict[str, list[tuple[str, str]]]) -> dict[str, list[EdgeId]]:
        """Gets all the edges that exist that are not in the toolkit files."""

        for edge_type in self.exclude_edges:
            edge_ids.pop(edge_type, None)

        spaces = set()
        for edges in edge_ids.values():
            for tk_edge in edges:
                spaces.add(tk_edge[0])

        types = [{"space": self.type_space, "externalId": xid} for xid in edge_ids.keys()]

        type_filter = In(("edge", "type"), types)
        existing_edges = self.client.data_modeling.instances.list(  # type: ignore[call-overload]
            instance_type="edge", limit=None, space=spaces, include_typing=True, filter=type_filter
        )

        types_with_nodes_to_delete: dict[str, list[EdgeId]] = {}
        for edge in existing_edges:
            type_xid = edge.type.external_id
            xids_for_type = edge_ids[type_xid]
            if (edge.space, edge.external_id) not in xids_for_type:
                if type_xid in types_with_nodes_to_delete.keys():
                    temp_list = types_with_nodes_to_delete[type_xid]
                    temp_list.append(edge.as_id())
                    types_with_nodes_to_delete[type_xid] = temp_list
                else:
                    types_with_nodes_to_delete[type_xid] = [edge.as_id()]

        return types_with_nodes_to_delete



from __future__ import annotations

import datetime
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, ClassVar, Literal, Optional, Union

from cognite.client import data_modeling as dm, CogniteClient
from pydantic import Field
from pydantic import field_validator, model_validator, ValidationInfo

from cognite.powerops.client._generated.config import global_config
from cognite.powerops.client._generated.data_classes._core import (
    DEFAULT_INSTANCE_SPACE,
    DEFAULT_QUERY_LIMIT,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelation,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
    T_DomainModelList,
    as_node_id,
    as_read_args,
    as_write_args,
    is_tuple_id,
    as_instance_dict_id,
    parse_single_connection,
    QueryCore,
    NodeQueryCore,
    StringFilter,
    ViewPropertyId,
    DateFilter,
    DirectRelationFilter,
    IntFilter,
)
from cognite.powerops.client._generated.data_classes._function_input import FunctionInput, FunctionInputWrite
if TYPE_CHECKING:
    from cognite.powerops.client._generated.data_classes._bid_configuration_day_ahead import BidConfigurationDayAhead, BidConfigurationDayAheadList, BidConfigurationDayAheadGraphQL, BidConfigurationDayAheadWrite, BidConfigurationDayAheadWriteList


__all__ = [
    "TaskDispatcherInput",
    "TaskDispatcherInputWrite",
    "TaskDispatcherInputList",
    "TaskDispatcherInputWriteList",
    "TaskDispatcherInputFields",
    "TaskDispatcherInputTextFields",
    "TaskDispatcherInputGraphQL",
]


TaskDispatcherInputTextFields = Literal["external_id", "workflow_execution_id", "function_name", "function_call_id"]
TaskDispatcherInputFields = Literal["external_id", "workflow_execution_id", "workflow_step", "function_name", "function_call_id", "bid_date"]

_TASKDISPATCHERINPUT_PROPERTIES_BY_FIELD = {
    "external_id": "externalId",
    "workflow_execution_id": "workflowExecutionId",
    "workflow_step": "workflowStep",
    "function_name": "functionName",
    "function_call_id": "functionCallId",
    "bid_date": "bidDate",
}


class TaskDispatcherInputGraphQL(GraphQLCore):
    """This represents the reading version of task dispatcher input, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the task dispatcher input.
        data_record: The data record of the task dispatcher input node.
        workflow_execution_id: The process associated with the function execution
        workflow_step: This is the step in the process.
        function_name: The name of the function
        function_call_id: The function call id
        bid_configuration: The bid configuration field.
        bid_date: The bid date
    """

    view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "TaskDispatcherInput", "1")
    workflow_execution_id: Optional[str] = Field(None, alias="workflowExecutionId")
    workflow_step: Optional[int] = Field(None, alias="workflowStep")
    function_name: Optional[str] = Field(None, alias="functionName")
    function_call_id: Optional[str] = Field(None, alias="functionCallId")
    bid_configuration: Optional[BidConfigurationDayAheadGraphQL] = Field(default=None, repr=False, alias="bidConfiguration")
    bid_date: Optional[datetime.date] = Field(None, alias="bidDate")

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values


    @field_validator("bid_configuration", mode="before")
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    def as_read(self) -> TaskDispatcherInput:
        """Convert this GraphQL format of task dispatcher input to the reading format."""
        return TaskDispatcherInput.model_validate(as_read_args(self))

    def as_write(self) -> TaskDispatcherInputWrite:
        """Convert this GraphQL format of task dispatcher input to the writing format."""
        return TaskDispatcherInputWrite.model_validate(as_write_args(self))


class TaskDispatcherInput(FunctionInput):
    """This represents the reading version of task dispatcher input.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the task dispatcher input.
        data_record: The data record of the task dispatcher input node.
        workflow_execution_id: The process associated with the function execution
        workflow_step: This is the step in the process.
        function_name: The name of the function
        function_call_id: The function call id
        bid_configuration: The bid configuration field.
        bid_date: The bid date
    """

    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "TaskDispatcherInput", "1")

    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("power_ops_types", "TaskDispatcherInput")
    bid_configuration: Union[BidConfigurationDayAhead, str, dm.NodeId, None] = Field(default=None, repr=False, alias="bidConfiguration")
    bid_date: Optional[datetime.date] = Field(None, alias="bidDate")
    @field_validator("bid_configuration", mode="before")
    @classmethod
    def parse_single(cls, value: Any, info: ValidationInfo) -> Any:
        return parse_single_connection(value, info.field_name)


    def as_write(self) -> TaskDispatcherInputWrite:
        """Convert this read version of task dispatcher input to the writing version."""
        return TaskDispatcherInputWrite.model_validate(as_write_args(self))



class TaskDispatcherInputWrite(FunctionInputWrite):
    """This represents the writing version of task dispatcher input.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the task dispatcher input.
        data_record: The data record of the task dispatcher input node.
        workflow_execution_id: The process associated with the function execution
        workflow_step: This is the step in the process.
        function_name: The name of the function
        function_call_id: The function call id
        bid_configuration: The bid configuration field.
        bid_date: The bid date
    """
    _container_fields: ClassVar[tuple[str, ...]] = ("bid_configuration", "bid_date", "function_call_id", "function_name", "workflow_execution_id", "workflow_step",)
    _direct_relations: ClassVar[tuple[str, ...]] = ("bid_configuration",)

    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "TaskDispatcherInput", "1")

    node_type: Union[dm.DirectRelationReference, dm.NodeId, tuple[str, str], None] = dm.DirectRelationReference("power_ops_types", "TaskDispatcherInput")
    bid_configuration: Union[BidConfigurationDayAheadWrite, str, dm.NodeId, None] = Field(default=None, repr=False, alias="bidConfiguration")
    bid_date: Optional[datetime.date] = Field(None, alias="bidDate")

    @field_validator("bid_configuration", mode="before")
    def as_node_id(cls, value: Any) -> Any:
        if isinstance(value, dm.DirectRelationReference):
            return dm.NodeId(value.space, value.external_id)
        elif isinstance(value, tuple) and len(value) == 2 and all(isinstance(item, str) for item in value):
            return dm.NodeId(value[0], value[1])
        elif isinstance(value, list):
            return [cls.as_node_id(item) for item in value]
        return value


class TaskDispatcherInputList(DomainModelList[TaskDispatcherInput]):
    """List of task dispatcher inputs in the read version."""

    _INSTANCE = TaskDispatcherInput
    def as_write(self) -> TaskDispatcherInputWriteList:
        """Convert these read versions of task dispatcher input to the writing versions."""
        return TaskDispatcherInputWriteList([node.as_write() for node in self.data])


    @property
    def bid_configuration(self) -> BidConfigurationDayAheadList:
        from ._bid_configuration_day_ahead import BidConfigurationDayAhead, BidConfigurationDayAheadList
        return BidConfigurationDayAheadList([item.bid_configuration for item in self.data if isinstance(item.bid_configuration, BidConfigurationDayAhead)])

class TaskDispatcherInputWriteList(DomainModelWriteList[TaskDispatcherInputWrite]):
    """List of task dispatcher inputs in the writing version."""

    _INSTANCE = TaskDispatcherInputWrite
    @property
    def bid_configuration(self) -> BidConfigurationDayAheadWriteList:
        from ._bid_configuration_day_ahead import BidConfigurationDayAheadWrite, BidConfigurationDayAheadWriteList
        return BidConfigurationDayAheadWriteList([item.bid_configuration for item in self.data if isinstance(item.bid_configuration, BidConfigurationDayAheadWrite)])


def _create_task_dispatcher_input_filter(
    view_id: dm.ViewId,
    workflow_execution_id: str | list[str] | None = None,
    workflow_execution_id_prefix: str | None = None,
    min_workflow_step: int | None = None,
    max_workflow_step: int | None = None,
    function_name: str | list[str] | None = None,
    function_name_prefix: str | None = None,
    function_call_id: str | list[str] | None = None,
    function_call_id_prefix: str | None = None,
    bid_configuration: str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference | Sequence[str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference] | None = None,
    min_bid_date: datetime.date | None = None,
    max_bid_date: datetime.date | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters: list[dm.Filter] = []
    if isinstance(workflow_execution_id, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("workflowExecutionId"), value=workflow_execution_id))
    if workflow_execution_id and isinstance(workflow_execution_id, list):
        filters.append(dm.filters.In(view_id.as_property_ref("workflowExecutionId"), values=workflow_execution_id))
    if workflow_execution_id_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("workflowExecutionId"), value=workflow_execution_id_prefix))
    if min_workflow_step is not None or max_workflow_step is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("workflowStep"), gte=min_workflow_step, lte=max_workflow_step))
    if isinstance(function_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("functionName"), value=function_name))
    if function_name and isinstance(function_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("functionName"), values=function_name))
    if function_name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("functionName"), value=function_name_prefix))
    if isinstance(function_call_id, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("functionCallId"), value=function_call_id))
    if function_call_id and isinstance(function_call_id, list):
        filters.append(dm.filters.In(view_id.as_property_ref("functionCallId"), values=function_call_id))
    if function_call_id_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("functionCallId"), value=function_call_id_prefix))
    if isinstance(bid_configuration, str | dm.NodeId | dm.DirectRelationReference) or is_tuple_id(bid_configuration):
        filters.append(dm.filters.Equals(view_id.as_property_ref("bidConfiguration"), value=as_instance_dict_id(bid_configuration)))
    if bid_configuration and isinstance(bid_configuration, Sequence) and not isinstance(bid_configuration, str) and not is_tuple_id(bid_configuration):
        filters.append(dm.filters.In(view_id.as_property_ref("bidConfiguration"), values=[as_instance_dict_id(item) for item in bid_configuration]))
    if min_bid_date is not None or max_bid_date is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("bidDate"), gte=min_bid_date.isoformat() if min_bid_date else None, lte=max_bid_date.isoformat() if max_bid_date else None))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None


class _TaskDispatcherInputQuery(NodeQueryCore[T_DomainModelList, TaskDispatcherInputList]):
    _view_id = TaskDispatcherInput._view_id
    _result_cls = TaskDispatcherInput
    _result_list_cls_end = TaskDispatcherInputList

    def __init__(
        self,
        created_types: set[type],
        creation_path: list[QueryCore],
        client: CogniteClient,
        result_list_cls: type[T_DomainModelList],
        expression: dm.query.NodeOrEdgeResultSetExpression | None = None,
        connection_name: str | None = None,
        connection_property: ViewPropertyId | None = None,
        connection_type: Literal["reverse-list"] | None = None,
        reverse_expression: dm.query.NodeOrEdgeResultSetExpression | None = None,
    ):
        from ._bid_configuration_day_ahead import _BidConfigurationDayAheadQuery

        super().__init__(
            created_types,
            creation_path,
            client,
            result_list_cls,
            expression,
            dm.filters.HasData(views=[self._view_id]),
            connection_name,
            connection_property,
            connection_type,
            reverse_expression,
        )

        if _BidConfigurationDayAheadQuery not in created_types and len(creation_path) + 1 < global_config.max_select_depth:
            self.bid_configuration = _BidConfigurationDayAheadQuery(
                created_types.copy(),
                self._creation_path,
                client,
                result_list_cls,
                dm.query.NodeResultSetExpression(
                    through=self._view_id.as_property_ref("bidConfiguration"),
                    direction="outwards",
                ),
                connection_name="bid_configuration",
                connection_property=ViewPropertyId(self._view_id, "bidConfiguration"),
            )

        self.space = StringFilter(self, ["node", "space"])
        self.external_id = StringFilter(self, ["node", "externalId"])
        self.workflow_execution_id = StringFilter(self, self._view_id.as_property_ref("workflowExecutionId"))
        self.workflow_step = IntFilter(self, self._view_id.as_property_ref("workflowStep"))
        self.function_name = StringFilter(self, self._view_id.as_property_ref("functionName"))
        self.function_call_id = StringFilter(self, self._view_id.as_property_ref("functionCallId"))
        self.bid_configuration_filter = DirectRelationFilter(self, self._view_id.as_property_ref("bidConfiguration"))
        self.bid_date = DateFilter(self, self._view_id.as_property_ref("bidDate"))
        self._filter_classes.extend([
            self.space,
            self.external_id,
            self.workflow_execution_id,
            self.workflow_step,
            self.function_name,
            self.function_call_id,
            self.bid_configuration_filter,
            self.bid_date,
        ])

    def list_task_dispatcher_input(self, limit: int = DEFAULT_QUERY_LIMIT) -> TaskDispatcherInputList:
        return self._list(limit=limit)


class TaskDispatcherInputQuery(_TaskDispatcherInputQuery[TaskDispatcherInputList]):
    def __init__(self, client: CogniteClient):
        super().__init__(set(), [], client, TaskDispatcherInputList)



from __future__ import annotations

import datetime
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, ClassVar, Literal, Optional, Union

from cognite.client import data_modeling as dm, CogniteClient
from pydantic import Field
from pydantic import field_validator, model_validator, ValidationInfo

from cognite.powerops.client._generated.config import global_config
from cognite.powerops.client._generated.data_classes._core import (
    DEFAULT_INSTANCE_SPACE,
    DEFAULT_QUERY_LIMIT,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelation,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
    T_DomainModelList,
    as_node_id,
    as_read_args,
    as_write_args,
    is_tuple_id,
    as_instance_dict_id,
    parse_single_connection,
    QueryCore,
    NodeQueryCore,
    StringFilter,
    ViewPropertyId,
    DirectRelationFilter,
    TimestampFilter,
)
if TYPE_CHECKING:
    from cognite.powerops.client._generated.data_classes._shop_file import ShopFile, ShopFileList, ShopFileGraphQL, ShopFileWrite, ShopFileWriteList
    from cognite.powerops.client._generated.data_classes._shop_scenario import ShopScenario, ShopScenarioList, ShopScenarioGraphQL, ShopScenarioWrite, ShopScenarioWriteList


__all__ = [
    "ShopCase",
    "ShopCaseWrite",
    "ShopCaseList",
    "ShopCaseWriteList",
    "ShopCaseFields",
    "ShopCaseGraphQL",
]


ShopCaseTextFields = Literal["external_id", ]
ShopCaseFields = Literal["external_id", "start_time", "end_time", "status"]

_SHOPCASE_PROPERTIES_BY_FIELD = {
    "external_id": "externalId",
    "start_time": "startTime",
    "end_time": "endTime",
    "status": "status",
}


class ShopCaseGraphQL(GraphQLCore):
    """This represents the reading version of shop case, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop case.
        data_record: The data record of the shop case node.
        scenario: The Shop scenario that was used to produce this result
        start_time: The start time of the case
        end_time: The end time of the case
        status: The status of the ShopCase
        shop_files: The list of shop files that are used in a shop run. This encompasses all shop files such as case,
            module series, cut files etc.
    """

    view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "ShopCase", "1")
    scenario: Optional[ShopScenarioGraphQL] = Field(default=None, repr=False)
    start_time: Optional[datetime.datetime] = Field(None, alias="startTime")
    end_time: Optional[datetime.datetime] = Field(None, alias="endTime")
    status: Optional[Literal["completed", "default", "failed", "notSet", "queued", "running", "stale", "timedOut", "triggered"]] = None
    shop_files: Optional[list[ShopFileGraphQL]] = Field(default=None, repr=False, alias="shopFiles")

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values


    @field_validator("scenario", "shop_files", mode="before")
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    def as_read(self) -> ShopCase:
        """Convert this GraphQL format of shop case to the reading format."""
        return ShopCase.model_validate(as_read_args(self))

    def as_write(self) -> ShopCaseWrite:
        """Convert this GraphQL format of shop case to the writing format."""
        return ShopCaseWrite.model_validate(as_write_args(self))


class ShopCase(DomainModel):
    """This represents the reading version of shop case.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop case.
        data_record: The data record of the shop case node.
        scenario: The Shop scenario that was used to produce this result
        start_time: The start time of the case
        end_time: The end time of the case
        status: The status of the ShopCase
        shop_files: The list of shop files that are used in a shop run. This encompasses all shop files such as case,
            module series, cut files etc.
    """

    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "ShopCase", "1")

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = None
    scenario: Union[ShopScenario, str, dm.NodeId, None] = Field(default=None, repr=False)
    start_time: Optional[datetime.datetime] = Field(None, alias="startTime")
    end_time: Optional[datetime.datetime] = Field(None, alias="endTime")
    status: Optional[Literal["completed", "default", "failed", "notSet", "queued", "running", "stale", "timedOut", "triggered"]] | str = None
    shop_files: Optional[list[Union[ShopFile, str, dm.NodeId]]] = Field(default=None, repr=False, alias="shopFiles")
    @field_validator("scenario", mode="before")
    @classmethod
    def parse_single(cls, value: Any, info: ValidationInfo) -> Any:
        return parse_single_connection(value, info.field_name)

    @field_validator("shop_files", mode="before")
    @classmethod
    def parse_list(cls, value: Any, info: ValidationInfo) -> Any:
        if value is None:
            return None
        return [parse_single_connection(item, info.field_name) for item in value]

    def as_write(self) -> ShopCaseWrite:
        """Convert this read version of shop case to the writing version."""
        return ShopCaseWrite.model_validate(as_write_args(self))



class ShopCaseWrite(DomainModelWrite):
    """This represents the writing version of shop case.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop case.
        data_record: The data record of the shop case node.
        scenario: The Shop scenario that was used to produce this result
        start_time: The start time of the case
        end_time: The end time of the case
        status: The status of the ShopCase
        shop_files: The list of shop files that are used in a shop run. This encompasses all shop files such as case,
            module series, cut files etc.
    """
    _container_fields: ClassVar[tuple[str, ...]] = ("end_time", "scenario", "start_time", "status",)
    _outwards_edges: ClassVar[tuple[tuple[str, dm.DirectRelationReference], ...]] = (("shop_files", dm.DirectRelationReference("power_ops_types", "ShopCase.shopFiles")),)
    _direct_relations: ClassVar[tuple[str, ...]] = ("scenario",)

    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "ShopCase", "1")

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, dm.NodeId, tuple[str, str], None] = None
    scenario: Union[ShopScenarioWrite, str, dm.NodeId, None] = Field(default=None, repr=False)
    start_time: Optional[datetime.datetime] = Field(None, alias="startTime")
    end_time: Optional[datetime.datetime] = Field(None, alias="endTime")
    status: Optional[Literal["completed", "default", "failed", "notSet", "queued", "running", "stale", "timedOut", "triggered"]] = None
    shop_files: Optional[list[Union[ShopFileWrite, str, dm.NodeId]]] = Field(default=None, repr=False, alias="shopFiles")

    @field_validator("scenario", "shop_files", mode="before")
    def as_node_id(cls, value: Any) -> Any:
        if isinstance(value, dm.DirectRelationReference):
            return dm.NodeId(value.space, value.external_id)
        elif isinstance(value, tuple) and len(value) == 2 and all(isinstance(item, str) for item in value):
            return dm.NodeId(value[0], value[1])
        elif isinstance(value, list):
            return [cls.as_node_id(item) for item in value]
        return value


class ShopCaseList(DomainModelList[ShopCase]):
    """List of shop cases in the read version."""

    _INSTANCE = ShopCase
    def as_write(self) -> ShopCaseWriteList:
        """Convert these read versions of shop case to the writing versions."""
        return ShopCaseWriteList([node.as_write() for node in self.data])


    @property
    def scenario(self) -> ShopScenarioList:
        from ._shop_scenario import ShopScenario, ShopScenarioList
        return ShopScenarioList([item.scenario for item in self.data if isinstance(item.scenario, ShopScenario)])
    @property
    def shop_files(self) -> ShopFileList:
        from ._shop_file import ShopFile, ShopFileList
        return ShopFileList([item for items in self.data for item in items.shop_files or [] if isinstance(item, ShopFile)])


class ShopCaseWriteList(DomainModelWriteList[ShopCaseWrite]):
    """List of shop cases in the writing version."""

    _INSTANCE = ShopCaseWrite
    @property
    def scenario(self) -> ShopScenarioWriteList:
        from ._shop_scenario import ShopScenarioWrite, ShopScenarioWriteList
        return ShopScenarioWriteList([item.scenario for item in self.data if isinstance(item.scenario, ShopScenarioWrite)])
    @property
    def shop_files(self) -> ShopFileWriteList:
        from ._shop_file import ShopFileWrite, ShopFileWriteList
        return ShopFileWriteList([item for items in self.data for item in items.shop_files or [] if isinstance(item, ShopFileWrite)])



def _create_shop_case_filter(
    view_id: dm.ViewId,
    scenario: str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference | Sequence[str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference] | None = None,
    min_start_time: datetime.datetime | None = None,
    max_start_time: datetime.datetime | None = None,
    min_end_time: datetime.datetime | None = None,
    max_end_time: datetime.datetime | None = None,
    status: Literal["completed", "default", "failed", "notSet", "queued", "running", "stale", "timedOut", "triggered"] | list[Literal["completed", "default", "failed", "notSet", "queued", "running", "stale", "timedOut", "triggered"]] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters: list[dm.Filter] = []
    if isinstance(scenario, str | dm.NodeId | dm.DirectRelationReference) or is_tuple_id(scenario):
        filters.append(dm.filters.Equals(view_id.as_property_ref("scenario"), value=as_instance_dict_id(scenario)))
    if scenario and isinstance(scenario, Sequence) and not isinstance(scenario, str) and not is_tuple_id(scenario):
        filters.append(dm.filters.In(view_id.as_property_ref("scenario"), values=[as_instance_dict_id(item) for item in scenario]))
    if min_start_time is not None or max_start_time is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("startTime"), gte=min_start_time.isoformat(timespec="milliseconds") if min_start_time else None, lte=max_start_time.isoformat(timespec="milliseconds") if max_start_time else None))
    if min_end_time is not None or max_end_time is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("endTime"), gte=min_end_time.isoformat(timespec="milliseconds") if min_end_time else None, lte=max_end_time.isoformat(timespec="milliseconds") if max_end_time else None))
    if isinstance(status, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("status"), value=status))
    if status and isinstance(status, list):
        filters.append(dm.filters.In(view_id.as_property_ref("status"), values=status))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None


class _ShopCaseQuery(NodeQueryCore[T_DomainModelList, ShopCaseList]):
    _view_id = ShopCase._view_id
    _result_cls = ShopCase
    _result_list_cls_end = ShopCaseList

    def __init__(
        self,
        created_types: set[type],
        creation_path: list[QueryCore],
        client: CogniteClient,
        result_list_cls: type[T_DomainModelList],
        expression: dm.query.NodeOrEdgeResultSetExpression | None = None,
        connection_name: str | None = None,
        connection_property: ViewPropertyId | None = None,
        connection_type: Literal["reverse-list"] | None = None,
        reverse_expression: dm.query.NodeOrEdgeResultSetExpression | None = None,
    ):
        from ._shop_file import _ShopFileQuery
        from ._shop_scenario import _ShopScenarioQuery

        super().__init__(
            created_types,
            creation_path,
            client,
            result_list_cls,
            expression,
            dm.filters.HasData(views=[self._view_id]),
            connection_name,
            connection_property,
            connection_type,
            reverse_expression,
        )

        if _ShopScenarioQuery not in created_types and len(creation_path) + 1 < global_config.max_select_depth:
            self.scenario = _ShopScenarioQuery(
                created_types.copy(),
                self._creation_path,
                client,
                result_list_cls,
                dm.query.NodeResultSetExpression(
                    through=self._view_id.as_property_ref("scenario"),
                    direction="outwards",
                ),
                connection_name="scenario",
                connection_property=ViewPropertyId(self._view_id, "scenario"),
            )

        if _ShopFileQuery not in created_types and len(creation_path) + 1 < global_config.max_select_depth:
            self.shop_files = _ShopFileQuery(
                created_types.copy(),
                self._creation_path,
                client,
                result_list_cls,
                dm.query.EdgeResultSetExpression(
                    direction="outwards",
                    chain_to="destination",
                ),
                connection_name="shop_files",
                connection_property=ViewPropertyId(self._view_id, "shopFiles"),
            )

        self.space = StringFilter(self, ["node", "space"])
        self.external_id = StringFilter(self, ["node", "externalId"])
        self.scenario_filter = DirectRelationFilter(self, self._view_id.as_property_ref("scenario"))
        self.start_time = TimestampFilter(self, self._view_id.as_property_ref("startTime"))
        self.end_time = TimestampFilter(self, self._view_id.as_property_ref("endTime"))
        self._filter_classes.extend([
            self.space,
            self.external_id,
            self.scenario_filter,
            self.start_time,
            self.end_time,
        ])

    def list_shop_case(self, limit: int = DEFAULT_QUERY_LIMIT) -> ShopCaseList:
        return self._list(limit=limit)


class ShopCaseQuery(_ShopCaseQuery[ShopCaseList]):
    def __init__(self, client: CogniteClient):
        super().__init__(set(), [], client, ShopCaseList)
